"""
Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE

Contains large portions of:
https://github.com/ValvePython/steam/blob/master/steam/core/cm.py
https://github.com/Rapptz/discord.py/tree/master/discord/gateway.py
The appropriate licenses are in LICENSE
"""

from __future__ import annotations

import asyncio
import concurrent.futures
import logging
import random
import struct
import sys
import threading
import time
import traceback
from collections.abc import AsyncGenerator, Callable
from dataclasses import dataclass
from datetime import timedelta
from gzip import FCOMMENT, FEXTRA, FHCRC, FNAME  # type: ignore
from operator import attrgetter
from typing import TYPE_CHECKING, Any, Generic, TypeVar, overload
from zlib import MAX_WBITS, decompress

import aiohttp
import async_timeout
import attr
from typing_extensions import TypeAlias

from . import utils
from ._const import DEFAULT_CMS
from .enums import IntEnum, PersonaState, PersonaStateFlag, Result
from .errors import NoCMsFound
from .iterators import AsyncIterator
from .models import Registerable, register, return_true
from .protobufs import EMsg, GCMsg, GCMsgProto, Msg, MsgProto, login
from .protobufs.client_server_2 import CMsgGcClient

if TYPE_CHECKING:
    from .client import Client
    from .enums import UIMode
    from .protobufs.base import CMsgMulti
    from .state import ConnectionState
    from .types.game import GameToDict
    from .types.http import Coro


__all__ = (
    "ConnectionClosed",
    "CMServerList",
    "SteamWebSocket",
    "Msgs",
)

log = logging.getLogger(__name__)
ProtoMsgs: TypeAlias = "MsgProto[Any] | Msg[Any]"
GCMsgs: TypeAlias = "GCMsgProto[Any] | GCMsg[Any]"
GCMsgsT = TypeVar("GCMsgsT", GCMsgProto[Any], GCMsg[Any])
ProtoMsgsT = TypeVar("ProtoMsgsT", MsgProto[Any], Msg[Any])
Msgs: TypeAlias = "ProtoMsgs | GCMsgs"
MsgsT = TypeVar("MsgsT", MsgProto[Any], Msg[Any], GCMsgProto[Any], GCMsg[Any])
M = TypeVar("M", bound=MsgProto[Any])
MsgProtoT = TypeVar("MsgProtoT", bound=MsgProto[Any])
MsgT = TypeVar("MsgT", bound=Msg[Any])
GCMsgT = TypeVar("GCMsgT", bound=GCMsg[Any])
GCMsgProtoT = TypeVar("GCMsgProtoT", bound=GCMsgProto[Any])

READ_U32 = struct.Struct("<I").unpack_from


@dataclass
class EventListener(Generic[MsgsT]):
    __slots__ = ("emsg", "check", "future")

    emsg: IntEnum | None
    check: Callable[[MsgsT], bool]
    future: asyncio.Future[MsgsT]


if not TYPE_CHECKING:
    EventListener.__class_getitem__ = classmethod(lambda cls, params: cls)


@attr.dataclass(slots=True)
class CMServer:
    _state: ConnectionState
    url: str
    weighted_load: float

    def connect(self) -> Coro[aiohttp.ClientWebSocketResponse]:
        return self._state.http.connect_to_cm(self.url)

    async def ping(self) -> float:
        try:
            async with async_timeout.timeout(5):
                async with self._state.http._session.get(f"https://{self.url}/cmping/") as resp:
                    if resp.status != 200:
                        raise KeyError
                    return int(resp.headers["X-Steam-CMLoad"])
        except (KeyError, asyncio.TimeoutError, aiohttp.ClientError):
            return float("inf")


class ConnectionClosed(Exception):
    def __init__(self, cm: CMServer, cms: CMServerList):
        self.cm = cm
        self.cm_list = cms
        super().__init__(f"Connection to {self.cm.url}, has closed.")


class WebSocketClosure(Exception):
    """An exception to make up for the fact that aiohttp doesn't signal closure."""


class CMServerList(AsyncIterator[CMServer]):
    def __init__(self, state: ConnectionState, first_cm_to_try: CMServer | None = None):
        super().__init__(state)
        self.cell_id = 0
        self._first_cm = first_cm_to_try

    async def fill(self, cell_id: int = 0) -> AsyncGenerator[CMServer, None]:
        if self._first_cm is not None:
            yield self._first_cm
        log.debug("Attempting to fetch servers from the WebAPI")
        self.cell_id = cell_id
        try:
            data = await self._state.http.get_cm_list(cell_id)
        except Exception:
            servers = [CMServer(self._state, url=cm_url, weighted_load=0) for cm_url in DEFAULT_CMS]
            random.shuffle(servers)
            log.debug("Error occurred when fetching CM server list, falling back to internal list", exc_info=True)
            for server in servers:
                yield server
            return

        resp = data["response"]
        if not resp["success"]:
            servers = [CMServer(self._state, url=cm_url, weighted_load=0) for cm_url in DEFAULT_CMS]
            random.shuffle(servers)
            log.debug("Error occurred when fetching CM server list, falling back to internal list", exc_info=True)
            for server in servers:
                yield server
            return

        hosts: list[dict[str, Any]] = resp["serverlist"]
        log.debug(f"Received {len(hosts)} servers from WebAPI")
        servers = [CMServer(self._state, url=server["endpoint"], weighted_load=server["wtd_load"]) for server in hosts]
        for server in sorted(servers, key=attrgetter("weighted_load")):  # they should already be sorted but oh well
            yield server


class KeepAliveHandler(threading.Thread):
    def __init__(self, ws: SteamWebSocket, interval: int):
        super().__init__()
        self.ws = ws
        self.interval = interval
        self._main_thread_id = self.ws.thread_id
        self.heartbeat = MsgProto[login.CMsgClientHeartBeat](EMsg.ClientHeartBeat, send_reply=True)
        self.msg = "Keeping websocket alive with heartbeat %s."
        self.block_msg = "Heartbeat blocked for more than {total} seconds."
        self.behind_msg = "Can't keep up, websocket is {total:.1f}s behind."
        self._stop_ev = threading.Event()
        self._last_ack = time.perf_counter()
        self._last_send = time.perf_counter()
        self.latency = float("inf")

    def run(self) -> None:
        while not self._stop_ev.wait(self.interval):
            if self._last_recv + 60 < time.perf_counter():
                log.warning("CM %r has stopped responding to the gateway. Closing and restarting.", self.ws.cm)
                coro = self.ws.close(4000)
                f = asyncio.run_coroutine_threadsafe(coro, loop=self.ws.loop)

                try:
                    f.result()
                except Exception:
                    log.exception("An error occurred while stopping the gateway. Ignoring.")
                finally:
                    return self.stop()

            log.debug(self.msg, self.heartbeat)
            coro = self.ws.send_proto(self.heartbeat)
            f = asyncio.run_coroutine_threadsafe(coro, loop=self.ws.loop)
            # block until sending is complete
            total = 0
            try:
                while True:
                    try:
                        f.result(timeout=10)
                        break
                    except concurrent.futures.TimeoutError:
                        total += 10
                        try:
                            frame = sys._current_frames()[self._main_thread_id]  # noqa
                        except KeyError:
                            msg = self.block_msg
                        else:
                            stack = "".join(traceback.format_stack(frame))
                            msg = f"{self.block_msg}\nLoop thread traceback (most recent call last):\n{stack}"
                        log.warning(msg.format(total=total))

            except Exception:
                self.stop()
            else:
                self._last_send = time.perf_counter()

    def stop(self) -> None:
        self._stop_ev.set()

    def tick(self) -> None:
        self._last_recv = time.perf_counter()

    def ack(self) -> None:
        ack_time = time.perf_counter()
        self._last_ack = ack_time
        self.latency = ack_time - self._last_send
        if self.latency > 10:
            log.warning(self.behind_msg.format(total=self.latency))


class SteamWebSocket(Registerable):
    parsers: dict[EMsg, Callable[..., Any]]

    def __init__(
        self, state: ConnectionState, socket: aiohttp.ClientWebSocketResponse, cm_list: CMServerList, cm: CMServer
    ):
        self.socket = socket

        # state stuff
        self._connection = state
        self.cm_list = cm_list
        self.cm = cm
        # the keep alive
        self._keep_alive: KeepAliveHandler
        self._dispatch = state.dispatch
        self.thread_id = threading.get_ident()

        # ws related stuff
        self.listeners: list[EventListener[Any]] = []
        self.parsers.update(state.parsers)
        self.closed = False

        self.session_id = 0
        self.steam_id = state.user.id64
        self._current_job_id = 0
        self._gc_current_job_id = 0

    @property
    def latency(self) -> float:
        """Measures latency between a heartbeat send and the heartbeat interval in seconds."""
        return self._keep_alive.latency

    @overload
    def wait_for(self, emsg: EMsg | None, check: Callable[[MsgT], bool] = ...) -> asyncio.Future[MsgT]:
        ...

    @overload
    def wait_for(self, emsg: EMsg | None, check: Callable[[MsgProtoT], bool] = ...) -> asyncio.Future[MsgProtoT]:
        ...

    def wait_for(
        self, emsg: EMsg | None, check: Callable[[ProtoMsgsT], bool] = return_true
    ) -> asyncio.Future[ProtoMsgsT]:
        future: asyncio.Future[ProtoMsgsT] = self.loop.create_future()
        entry = EventListener(emsg=emsg, check=check, future=future)
        self.listeners.append(entry)
        return future

    @classmethod
    async def from_client(
        cls, client: Client, cm: CMServer | None = None, cm_list: CMServerList | None = None
    ) -> SteamWebSocket:
        state = client._connection
        cm_list = cm_list or CMServerList(state, cm)
        token = await client.token()
        async for cm in cm_list:
            log.info(f"Attempting to create a websocket connection to: {cm}")
            socket = await cm.connect()
            log.debug(f"Connected to {cm}")

            self = cls(state, socket, cm_list, cm)
            await self.send_proto(
                MsgProto(
                    EMsg.ClientLogon,
                    account_name=client.username,
                    web_logon_nonce=token,
                    client_os_type=4294966596,
                    client_language=self._connection.http.language.api_name,
                    protocol_version=65580,
                    chat_mode=2,
                    ui_mode=self._connection._ui_mode,
                    qos_level=2,
                )
            )
            self._dispatch("connect")
            return self
        raise NoCMsFound("No CMs found could be connected to. Steam is likely down")

    async def poll_event(self) -> None:
        try:
            message = await self.socket.receive()
            if message.type is aiohttp.WSMsgType.BINARY and message.data:  # it can sometimes be None/empty
                return self.receive(message.data)
            if message.type is aiohttp.WSMsgType.ERROR:
                log.debug(f"Received {message}")
                raise message.data
            if message.type in (aiohttp.WSMsgType.CLOSED, aiohttp.WSMsgType.CLOSE, aiohttp.WSMsgType.CLOSING):
                log.debug(f"Received {message}")
                raise WebSocketClosure
            log.debug(f"Dropped unexpected message type: {message}")
        except WebSocketClosure:
            await self.handle_close()

    def receive(self, message: bytes) -> None:
        (emsg_value,) = READ_U32(message)
        emsg = EMsg(utils.clear_proto_bit(emsg_value))

        try:
            msg = MsgProto(emsg, message) if utils.is_proto(emsg_value) else Msg(emsg, message, extended=True)
        except Exception as exc:
            return log.error(f"Failed to deserialize message: {emsg!r}, {message!r}", exc_info=exc)

        log.debug("Socket has received %r from the websocket.", msg)

        if hasattr(self, "_keep_alive"):
            self._keep_alive.tick()

        self._dispatch("socket_receive", msg)
        self.run_parser(emsg, msg)

        # remove the dispatched listener
        removed: list[int] = []
        for idx, entry in enumerate(self.listeners):
            if entry.emsg != emsg and entry.emsg is not None:
                continue

            future = entry.future
            if future.cancelled():
                removed.append(idx)
                continue

            try:
                valid = entry.check(msg)
            except Exception as exc:
                future.set_exception(exc)
                removed.append(idx)
            else:
                if valid:
                    future.set_result(msg)
                    removed.append(idx)

        for idx in reversed(removed):
            del self.listeners[idx]

    async def send(self, data: bytes) -> None:
        try:
            await self.socket.send_bytes(data)
        except ConnectionResetError:
            log.info("Connection closed")
            await self.handle_close()

    async def send_proto(self, message: ProtoMsgs) -> None:
        message.steam_id = self.steam_id
        message.session_id = self.session_id

        self._dispatch("socket_send", message)
        await self.send(bytes(message))

    async def send_gc_message(self, msg: GCMsgs) -> int:  # for ext's to send GC messages
        client = self._connection.client
        if __debug__ or TYPE_CHECKING:
            from .ext._gc import Client as GCClient

            assert isinstance(client, GCClient), "Attempting to send a GC message without a GC client"

        app_id = client._GAME.id
        message = MsgProto[CMsgGcClient](
            EMsg.ClientToGC,
            appid=app_id,
            msgtype=utils.set_proto_bit(msg.msg) if isinstance(msg, GCMsgProto) else msg.msg,
            payload=bytes(msg),
        )
        message.header.body.routing_app_id = app_id
        message.header.body.job_id_source = self._gc_current_job_id = (self._gc_current_job_id + 1) % 10000 or 1

        log.debug("Sending GC message %r", msg)
        self._dispatch("gc_message_send", msg)
        await self.send_proto(message)
        return message.header.body.job_id_source

    async def close(self, code: int = 1000) -> None:
        message = MsgProto(EMsg.ClientLogOff)
        message.steam_id = self.steam_id
        message.session_id = self.session_id
        await self.socket.close(code=code, message=bytes(message))

    @register(EMsg.ClientLoggedOff)
    async def handle_close(self, _: Any = None) -> None:
        if self.closed:  # don't want ConnectionClosed to be raised multiple times
            return
        if not self.socket.closed:
            await self.close()
            try:
                await self.cm_list.__anext__()  # pop the disconnected cm
            except StopAsyncIteration:
                pass
        if hasattr(self, "_keep_alive"):
            self._keep_alive.stop()
            del self._keep_alive
        log.info("Websocket closed, cannot reconnect.")
        self.closed = True
        raise ConnectionClosed(self.cm, self.cm_list)

    @register(EMsg.ClientLogOnResponse)
    async def handle_logon(self, msg: MsgProto[login.CMsgClientLogonResponse]) -> None:
        if msg.result != Result.OK:
            log.debug(f"Failed to login with result: {msg.result}")
            if msg.result == Result.InvalidPassword:
                http = self._connection.http
                await http.logout()
                await http.login(http.username, http.password, shared_secret=http.shared_secret)
            return await self.handle_close()

        self.session_id = msg.session_id
        self.cm_list.cell_id = msg.body.cell_id

        interval = msg.body.out_of_game_heartbeat_seconds
        self._keep_alive = KeepAliveHandler(ws=self, interval=interval)
        self._keep_alive.start()
        log.debug("Heartbeat started.")

        await self.send_um("ChatRoom.GetMyChatRoomGroups")
        await self.change_presence(
            games=self._connection._games,
            state=self._connection._state,
            flags=self._connection._flags,
            force_kick=self._connection._force_kick,
        )
        await self.send_proto(MsgProto(EMsg.ClientGetEmoticonList))
        await self.send_proto(MsgProto(EMsg.ClientRequestCommentNotifications))
        await self.send_proto(MsgProto(EMsg.ClientServerTimestampRequest, client_request_timestamp=time.time() * 1000))

        log.debug("Logon completed")

    @register(EMsg.ClientHeartBeat)
    def ack_heartbeat(self, msg: MsgProto[login.CMsgClientHeartBeat]) -> None:
        self._keep_alive.ack()

    @register(EMsg.ClientServerTimestampResponse)
    def set_steam_time(self, msg: MsgProto[login.CMsgClientServerTimestampResponse]) -> None:
        self.server_offset = timedelta(milliseconds=msg.body.server_timestamp_ms - msg.body.client_request_timestamp)

    @staticmethod
    def unpack_multi(msg: MsgProto[CMsgMulti]) -> bytes | None:
        data = msg.body.message_body
        log.debug(f"Decompressing payload ({len(data)} -> {msg.body.size_unzipped})")
        if data[:2] != b"\037\213":
            return log.info("Received a file that's not GZipped")

        flag = int.from_bytes(data[3:4], byteorder="little")
        position = 10

        if flag:  # this isn't ever hit, might as well save a few nanos
            if flag & FEXTRA:
                extra_len = int.from_bytes(data[position:2], byteorder="little")
                position += 2 + extra_len
            if flag & FNAME:
                while True:
                    terminator = data[position:1]
                    position += 1
                    if not terminator or terminator == b"\000":
                        break
            if flag & FCOMMENT:
                while True:
                    terminator = data[position:1]
                    position += 1
                    if not terminator or terminator == b"\000":
                        break
            if flag & FHCRC:
                position += 2

        decompressed = decompress(data[position:], wbits=-MAX_WBITS)

        if len(decompressed) != msg.body.size_unzipped:
            return log.info(f"Unzipped size mismatch for multi payload {msg}, discarding")

        return decompressed

    @register(EMsg.Multi)
    def handle_multi(self, msg: MsgProto[CMsgMulti]) -> None:
        log.debug("Received a multi")
        data = self.unpack_multi(msg) if msg.body.size_unzipped else msg.body.message_body

        while data:
            (size,) = READ_U32(data)
            self.receive(data[4 : 4 + size])
            data = data[4 + size :]

    @property
    def next_job_id(self) -> int:
        self._current_job_id = (self._current_job_id + 1) % 10000 or 1
        return self._current_job_id

    async def send_um(self, __name: str, **kwargs: Any) -> int:
        msg = MsgProto(EMsg.ServiceMethodCallFromClient, um_name=__name, **kwargs)
        msg.header.body.job_id_source = self.next_job_id
        await self.send_proto(msg)
        return msg.header.body.job_id_source

    async def send_um_and_wait(
        self,
        __name: str,
        check: Callable[[M], bool] | None = None,
        **kwargs: Any,
    ) -> M:
        job_id = await self.send_um(__name, **kwargs)
        check = check or (lambda msg: msg.header.body.job_id_target == job_id)
        return await self.wait_for(EMsg.ServiceMethodResponse, check=check)

    @overload
    async def send_proto_and_wait(self, msg: MsgT, check: Callable[[MsgT], bool] | None = ...) -> MsgT:
        ...

    @overload
    async def send_proto_and_wait(self, msg: MsgProtoT, check: Callable[[MsgProtoT], bool] | None = ...) -> MsgProtoT:
        ...

    async def send_proto_and_wait(
        self, msg: ProtoMsgs, check: Callable[[ProtoMsgsT], bool] | None = None
    ) -> ProtoMsgsT:
        msg.header.body.job_id_source = job_id = self.next_job_id
        await self.send_proto(msg)
        check = check or (lambda msg: msg.header.body.job_id_target == job_id)
        return await self.wait_for(None, check=check)

    async def change_presence(
        self,
        *,
        games: list[GameToDict] | None = None,
        state: PersonaState | None = None,
        flags: PersonaStateFlag | None = None,
        ui_mode: UIMode | None = None,
        force_kick: bool = False,
    ) -> None:
        self._connection._games = games or self._connection._games
        self._connection._state = state or self._connection._state
        self._connection._ui_mode = ui_mode or self._connection._ui_mode
        self._connection._flags = flags or self._connection._flags
        self._connection._force_kick = force_kick

        if force_kick:
            kick_msg = MsgProto(EMsg.ClientKickPlayingSession)
            log.debug("Kicking any currently playing sessions")
            await self.send_proto(kick_msg)
        if games:
            games_msg = MsgProto(EMsg.ClientGamesPlayedWithDataBlob, games_played=games)
            log.debug("Sending %r to change activity", games_msg)
            await self.send_proto(games_msg)
        if state is not None or flags is not None:
            state_msg = MsgProto(EMsg.ClientChangeStatus, persona_state=state, persona_state_flags=flags)
            log.debug("Sending %r to change state", state_msg)
            await self.send_proto(state_msg)
        if ui_mode is not None:
            ui_mode_msg = MsgProto(EMsg.ClientCurrentUIMode, uimode=ui_mode)
            log.debug("Sending %r to change UI mode", ui_mode_msg)
            await self.send_proto(ui_mode_msg)



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from collections.abc import AsyncGenerator, Mapping
from contextlib import asynccontextmanager
from dataclasses import dataclass
from datetime import datetime
from io import BytesIO
from typing import TYPE_CHECKING, cast

from typing_extensions import Final, Protocol
from yarl import URL

from .game import StatefulGame
from .models import _IOMixin
from .utils import DateTime

if TYPE_CHECKING:
    from .message import Authors, Message
    from .protobufs.friends import (
        CMsgClientEmoticonListEffect as ClientEffectProto,
        CMsgClientEmoticonListEmoticon as ClientEmoticonProto,
        CMsgClientEmoticonListSticker as ClientStickerProto,
    )
    from .state import ConnectionState


__all__ = (
    "Award",
    "AwardReaction",
    "PartialMessageReaction",
    "MessageReaction",
    "Emoticon",
    "Sticker",
    "ClientEmoticon",
    "ClientSticker",
)


class ReactionProtocol(Protocol):
    reactionid: int
    count: int


@dataclass
class _Reaction:
    __slots__ = tuple(ReactionProtocol.__annotations__)
    __annotations__ = ReactionProtocol.__annotations__


BASE_REACTION_URL: Final = "https://store.cloudflare.steamstatic.com/public/images/loyalty/reactions/{type}/{id}.png"
BASE_ECONOMY_URL: Final = URL("https://community.akamai.steamstatic.com/economy")
AWARD_ID_TO_NAME: Final = cast("Mapping[int, str]", {
    1: "Deep Thoughts",
    2: "Heartwarming",
    3: "Hilarious",
    4: "Hot Take",
    5: "Poetry",
    6: "Extra Helpful",
    7: "Gotta Have It",
    8: "Michelangelo",
    9: "Treasure",
    10: "Mind Blown",
    11: "Golden Unicorn",
    12: "Mad Scientist",
    13: "Clever",
    14: "Warm Blanket",
    15: "Saucy",
    16: "Slow Clap",
    17: "Take My Points",
    18: "Wholesome",
    19: "Jester",
    20: "Fancy Pants",
    21: "Whoa",
    22: "Super Star",
    23: "Wild",
})  # fmt: skip


class Award(_IOMixin):
    """Represents an award.

    Attributes
    ----------
    id
        The ID of the award.
    name
        The english localised name of the award.
    """

    def __init__(self, state: ConnectionState, id: int):
        self._state = state
        self.id = id
        self.name = AWARD_ID_TO_NAME.get(self.id, f"Unknown Award {self.id}")

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r} id={self.id}>"

    @property
    def url(self) -> str:
        """The reaction's icon URL."""
        return BASE_REACTION_URL.format(type="still", id=self.id)

    @property
    def animated_url(self) -> str:
        """The reaction's icon animated URL."""
        return BASE_REACTION_URL.format(type="animated", id=self.id)

    @asynccontextmanager
    async def open(self, *, animated: bool = False) -> AsyncGenerator[BytesIO, None]:
        async with self._state.http._session.get(self.animated_url if animated else self.url) as r:
            yield BytesIO(await r.read())


class AwardReaction:
    """Represents an award on user generated content.

    Attributes
    ----------
    award
        The award given to the user generated content.
    count
        The reactions count on the post.
    """

    def __init__(self, state: ConnectionState, proto: ReactionProtocol):
        self._state = state
        self.award = Award(state, proto.reactionid)
        self.count = proto.count

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} award={self.award!r} count={self.count}>"


@dataclass
class PartialMessageReaction:
    """Represents a partial reaction to a message."""

    _state: ConnectionState
    message: Message
    """The message the reaction applies to. """
    emoticon: Emoticon | None
    """The emoticon that was reacted with."""
    sticker: Sticker | None
    """The sticker that was reacted with."""


@dataclass
class MessageReaction(PartialMessageReaction):
    """Represents a reaction to a message."""

    user: Authors
    """The user that reacted to the message."""
    created_at: datetime | None = None
    """The time the reaction was added to the message."""
    ordinal: int | None = None
    """The ordinal of the the message."""

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented

        return (
            self.message == other.message
            and self.emoticon == other.emoticon
            and self.sticker == other.sticker
            and self.user == other.user
        )


class BaseEmoticon(_IOMixin):
    __slots__ = ("_state", "name")
    url: str

    def __init__(self, state: ConnectionState, name: str):
        self._state = state
        self.name = name.strip(":")  # :emoji_name:

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r}>"


class Emoticon(BaseEmoticon):
    """Represents an emoticon in chat.

    .. container:: operations

        .. describe:: str(x)

            The string to send this emoticon in chat.

    Attributes
    ----------
    name
        The name of this emoticon.
    """

    __slots__ = ()

    def __str__(self):
        return f":{self.name}:"

    @property
    def url(self) -> str:
        """The URL for this emoticon."""
        return str(BASE_ECONOMY_URL / "emoticonlarge" / self.name)

    async def game(self) -> StatefulGame:
        """Fetches this emoticon's associated game.

        Note
        ----
        This game has its :attr:`~Game.name` set unlike :meth:`Sticker.game`.
        """
        data = await self._state.http.get(BASE_ECONOMY_URL / "emoticonhoverjson" / self.name)
        return StatefulGame(self._state, id=data["appid"], name=data["app_name"])


class Sticker(BaseEmoticon):
    """Represents a sticker in chat.

    .. container:: operations

        .. describe:: str(x)

            The way to send this sticker in chat.

    Note
    ----
    Unlike :class:`Emoticon` this can only be sent in a message by itself.

    Attributes
    ----------
    name
        The name of this sticker.
    """

    __slots__ = ()

    def __str__(self) -> str:
        return f"/sticker {self.name}"

    @property
    def url(self) -> str:
        """The URL for this sticker."""
        return str(BASE_ECONOMY_URL / "sticker" / self.name)

    async def game(self) -> StatefulGame:
        """Fetches this sticker's associated game."""
        data = await self._state.http.get(BASE_ECONOMY_URL / "stickerjson" / self.name)
        return StatefulGame(self._state, id=data["appid"])


class BaseClientEmoticon(BaseEmoticon):
    __slots__ = ("count", "use_count", "last_used", "received_at")

    def __init__(self, state: ConnectionState, proto: ClientEmoticonProto | ClientStickerProto):
        super().__init__(state, proto.name)
        self.count = proto.count
        self.use_count = proto.use_count
        self.last_used = DateTime.from_timestamp(proto.time_last_used)
        self.received_at = DateTime.from_timestamp(proto.time_received)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r} count={self.count!r} use_count={self.use_count!r}>"


class ClientEmoticon(BaseClientEmoticon, Emoticon):
    """Represents an :class:`Emoticon` the :class:`ClientUser` owns."""

    __slots__ = ()


class ClientSticker(BaseClientEmoticon, Sticker):
    """Represents a :class:`Sticker` the :class:`ClientUser` owns."""

    __slots__ = ("_app_id",)

    def __init__(self, state: ConnectionState, proto: ClientStickerProto):
        super().__init__(state, proto)
        self._app_id = proto.appid

    async def game(self) -> StatefulGame:
        return StatefulGame(self._state, id=self._app_id)  # no point fetching



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from collections.abc import Coroutine
from typing import Any, Generic, TypedDict, TypeVar

from typing_extensions import TypeAlias
from yarl import URL

T = TypeVar("T")


class ResponseDict(TypedDict, Generic[T]):
    response: T


Coro: TypeAlias = "Coroutine[Any, Any, T]"
ResponseType: TypeAlias = "Coro[ResponseDict[T]]"
StrOrURL: TypeAlias = URL | str



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import asyncio
import copy
import logging
import re
import urllib.parse
import warnings
from base64 import b64encode
from collections.abc import Callable, Iterable
from datetime import datetime
from sys import version_info
from time import time
from typing import TYPE_CHECKING, Any, TypeVar

import aiohttp
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from yarl import URL as _URL

from . import errors, utils
from .__metadata__ import __version__
from ._const import JSON_DUMPS, JSON_LOADS, URL
from .abc import SteamID
from .enums import Language
from .guard import generate_one_time_code
from .models import PriceOverviewDict, api_route
from .user import ClientUser, _BaseUser
from .utils import cached_property

if TYPE_CHECKING:
    from .client import Client
    from .image import Image
    from .types import game, trade
    from .types.http import Coro, ResponseType, StrOrURL
    from .types.package import FetchedPackage
    from .types.user import User


T = TypeVar("T")
log = logging.getLogger(__name__)


async def json_or_text(r: aiohttp.ClientResponse, *, loads: Callable[[str], Any] = JSON_LOADS) -> Any:
    text = await r.text()
    try:
        if "application/json" in r.headers["Content-Type"]:
            return loads(text)
    except KeyError:
        pass
    return text


class HTTPClient:
    """The HTTP Client that interacts with the Steam web API."""

    def __init__(self, client: Client, **options: Any):
        self._session: aiohttp.ClientSession = None  # type: ignore  # filled in login
        self.user: ClientUser = None  # type: ignore
        self._client = client

        self.username: str
        self.password: str
        self.api_key: str | None = None
        self.shared_secret: str | None
        self.language: Language = options.get("language", Language.English)

        self._one_time_code = ""
        self._email_code = ""
        self._captcha_id = "-1"
        self._captcha_text = ""
        self._steam_id = ""

        self.logged_in = False
        self.user_agent = (
            f"steam.py/{__version__} client (https://github.com/Gobot1234/steam.py), "
            f"Python/{version_info.major}.{version_info.minor}, aiohttp/{aiohttp.__version__}"
        )

        self.proxy: str | None = options.get("proxy")
        self.proxy_auth: aiohttp.BasicAuth | None = options.get("proxy_auth")
        self.connector: aiohttp.BaseConnector | None = options.get("connector")
        self.personal_inventory_lock = asyncio.Lock()

    def clear(self) -> None:
        self._session = aiohttp.ClientSession(
            cookies={"Steam_Language": self.language.api_name},
            connector=self.connector,
            json_serialize=JSON_DUMPS,
        )

    async def request(self, method: str, url: StrOrURL, **kwargs: Any) -> Any:  # adapted from d.py
        kwargs["headers"] = {"User-Agent": self.user_agent, **kwargs.get("headers", {})}
        payload = kwargs.get("data")

        for tries in range(5):
            async with self._session.request(method, url, **kwargs, proxy=self.proxy, proxy_auth=self.proxy_auth) as r:
                log.debug("%s %s with PAYLOAD: %s has returned %d", method, r.url, payload, r.status)

                # even errors have text involved in them so this is safe to call
                data = await json_or_text(r)

                # the request was successful so just return the text/json
                if 200 <= r.status < 300:
                    log.debug("%s %s has received %s", method, r.url, data)
                    return data

                # we are being rate limited
                elif r.status == 429:
                    # I haven't been able to get any X-Retry-After headers from the API but we should probably still
                    # handle it
                    log.warning("We are being Rate limited")
                    try:
                        await asyncio.sleep(float(r.headers["X-Retry-After"]))
                    except KeyError:  # steam being un-helpful as usual
                        await asyncio.sleep(2**tries)
                    continue

                # we've received a 500 or 502, an unconditional retry
                elif r.status in {500, 502}:
                    await asyncio.sleep(1 + tries * 3)
                    continue

                # been logged out
                elif 300 <= r.status <= 399 and "login" in r.headers.get("location", ""):
                    log.debug("Logged out of session re-logging in")
                    await self.login(self.username, self.password, self.shared_secret)
                    continue

                elif r.status == 401:
                    if not data:
                        raise errors.HTTPException(r, data)
                    # api key either got revoked or it was never valid
                    if "Access is denied. Retrying will not help. Please verify your <pre>key=</pre>" in data:
                        # time to fetch a new key
                        key = await self.get_api_key()
                        assert key is not None
                        self.api_key = kwargs["key"] = key
                        # retry with our new key

                # the usual error cases
                elif r.status == 403:
                    raise errors.Forbidden(r, data)
                elif r.status == 404:
                    raise errors.NotFound(r, data)
                else:
                    raise errors.HTTPException(r, data)

        # we've run out of retries, raise
        raise errors.HTTPException(r, data)

    def get(self, url: StrOrURL, **kwargs: Any) -> Coro[Any]:
        return self.request("GET", url, **kwargs)

    def post(self, url: StrOrURL, **kwargs: Any) -> Coro[Any]:
        return self.request("POST", url, **kwargs)

    def connect_to_cm(self, cm: str) -> Coro[aiohttp.ClientWebSocketResponse]:
        headers = {"User-Agent": self.user_agent}
        return self._session.ws_connect(
            f"wss://{cm}/cmsocket/", headers=headers, proxy=self.proxy, proxy_auth=self.proxy_auth
        )

    async def login(self, username: str, password: str, shared_secret: str | None) -> None:
        self.username = username
        self.password = password
        self.shared_secret = shared_secret
        self.clear()

        try:
            resp = await self._send_login_request()

            if resp.get("captcha_needed") and resp.get("message") != "Please wait and try again later.":
                self._captcha_id = resp["captcha_gid"]
                print(
                    "Please enter the captcha text at "
                    f"https://steamcommunity.com/login/rendercaptcha/?gid={resp['captcha_gid']}"
                )
                captcha_text = await utils.ainput(">>> ")
                self._captcha_text = captcha_text.strip()
                return await self.login(username, password, shared_secret)

            if not resp["success"]:
                raise errors.InvalidCredentials(resp.get("message", "An unexpected error occurred"))

            jar = self._session.cookie_jar
            cookies = jar.filter_cookies(URL.COMMUNITY)
            for url in resp["transfer_urls"]:
                jar.update_cookies(copy.deepcopy(cookies), _URL(url).origin())

            self.api_key = await self.get_api_key()
            if self.api_key is None:
                log.info("Failed to get API key")

                async def get_user(user_id64: int) -> User:  # this is a lie
                    user_id = user_id64 & 0xFFFFFFFF
                    ret = await self.get(URL.COMMUNITY / f"miniprofile/{user_id}/json")
                    ret["steamid"] = user_id64
                    return ret

                async def get_users(user_id64s: Iterable[int]) -> list[User]:
                    return await asyncio.gather(*(self.get_user(user_id64) for user_id64 in user_id64s))  # type: ignore

                _BaseUser._patch_without_api()
                self.get_user = get_user
                self.get_users = get_users
                warnings.warn(
                    "Some methods of User objects are not available as no API key can be generated", UserWarning
                )
                await self.get(URL.COMMUNITY / "home")

            data = await self.get_user(resp["transfer_parameters"]["steamid"])
            state = self._client._connection
            assert data
            self.user = ClientUser(state=state, data=data)
            state._users[self.user.id64] = self.user
            self.logged_in = True
            self._client.dispatch("login")

        except:
            await self._session.close()
            raise

    @cached_property  # should always be called after making at least one request
    def session_id(self) -> str:
        cookies = self._session.cookie_jar.filter_cookies(URL.COMMUNITY)
        return cookies["sessionid"].value

    async def close(self) -> None:
        await self.logout()
        await self._session.close()

    async def logout(self) -> None:
        log.debug("Logging out of session")
        payload = {"sessionid": self.session_id}
        await self.post(URL.COMMUNITY / "login/logout", data=payload)
        self.logged_in = False
        self.user = None  # type: ignore
        self._client.dispatch("logout")

    async def _get_rsa_params(self) -> tuple[bytes, int]:
        payload = {"username": self.username, "donotcache": int(time() * 1000)}
        try:
            key_response = await self.post(URL.COMMUNITY / "login/getrsakey", data=payload)
        except Exception as exc:
            raise errors.LoginError("Could not obtain RSA key") from exc
        try:
            n = int(key_response["publickey_mod"], 16)
            e = int(key_response["publickey_exp"], 16)
            rsa_timestamp = key_response["timestamp"]
        except KeyError:
            raise errors.LoginError("Could not obtain rsa-key") from None
        else:
            return (
                b64encode(
                    rsa.RSAPublicNumbers(e, n).public_key().encrypt(self.password.encode("utf-8"), padding.PKCS1v15())
                ),
                rsa_timestamp,
            )

    async def _send_login_request(self) -> dict[str, Any]:
        password, timestamp = await self._get_rsa_params()
        payload = {
            "username": self.username,
            "password": password.decode(),
            "emailauth": self._email_code,
            "emailsteamid": self._steam_id,
            "twofactorcode": (
                generate_one_time_code(self.shared_secret) if self.shared_secret is not None else self._one_time_code
            ),
            # attempting this straight away makes login a bit faster for everyone with a shared_secret and doesn't
            # hurt performance for others
            "captchagid": self._captcha_id,
            "captcha_text": self._captcha_text,
            "loginfriendlyname": self.user_agent,
            "rsatimestamp": timestamp,
            "remember_login": True,
            "donotcache": int(time() * 1000),
        }
        try:
            resp = await self.post(URL.COMMUNITY / "login/dologin", data=payload)
            if resp.get("requires_twofactor"):
                self._one_time_code = await self._client.code()
            elif resp.get("emailauth_needed"):
                self._steam_id = resp.get("emailsteamid")
                self._email_code = await self._client.code()
            else:
                return resp
            return await self._send_login_request()
        except errors.LoginError:
            raise
        except Exception as exc:
            try:
                msg = exc.args[0]
            except IndexError:
                msg = None
            raise errors.LoginError(msg) from exc

    async def get_user(self, user_id64: int) -> User | None:
        params = {"key": self.api_key, "steamids": user_id64}
        resp = await self.get(api_route("ISteamUser/GetPlayerSummaries", version=2), params=params)
        return resp["response"]["players"][0] if resp["response"]["players"] else None

    async def get_users(self, user_id64s: Iterable[int]) -> list[User]:
        ret: list[User] = []

        for resp in await asyncio.gather(  # gather all the requests concurrently
            *(
                self.get(
                    api_route("ISteamUser/GetPlayerSummaries", version=2),
                    params={"key": self.api_key, "steamids": ",".join(map(str, sublist))},
                )
                for sublist in utils.chunk(user_id64s, 100)
            )
        ):
            ret.extend(resp["response"]["players"])
        return ret

    def add_user(self, user_id64: int) -> Coro[None]:
        payload = {
            "sessionID": self.session_id,
            "steamid": user_id64,
            "accept_invite": 0,
        }
        return self.post(URL.COMMUNITY / "actions/AddFriendAjax", data=payload)

    def remove_user(self, user_id64: int) -> Coro[None]:
        payload = {
            "sessionID": self.session_id,
            "steamid": user_id64,
        }
        return self.post(URL.COMMUNITY / "actions/RemoveFriendAjax", data=payload)

    def block_user(self, user_id64: int) -> Coro[None]:
        payload = {"sessionID": self.session_id, "steamid": user_id64, "block": 1}
        return self.post(URL.COMMUNITY / "actions/BlockUserAjax", data=payload)

    def unblock_user(self, user_id64: int) -> Coro[None]:
        payload = {"sessionID": self.session_id, "steamid": user_id64, "block": 0}
        return self.post(URL.COMMUNITY / "actions/BlockUserAjax", data=payload)

    def accept_user_invite(self, user_id64: int) -> Coro[None]:
        payload = {
            "sessionID": self.session_id,
            "steamid": user_id64,
            "accept_invite": 1,
        }
        return self.post(URL.COMMUNITY / "actions/AddFriendAjax", data=payload)

    def decline_user_invite(self, user_id64: int) -> Coro[None]:
        payload = {
            "sessionID": self.session_id,
            "steamid": user_id64,
            "accept_invite": 0,
        }
        return self.post(URL.COMMUNITY / "actions/IgnoreFriendInviteAjax", data=payload)

    def get_user_games(self, user_id64: int) -> ResponseType[game.GetOwnedGames]:
        params = {
            "key": self.api_key,
            "steamid": user_id64,
            "include_appinfo": 1,
            "include_played_free_games": 1,
        }
        return self.get(api_route("IPlayerService/GetOwnedGames"), params=params)

    def get_user_inventory(
        self, user_id64: int, app_id: int, context_id: int, language: Language | None
    ) -> Coro[trade.Inventory]:
        params = {
            "count": 5000,
            "l": (language or self.language).api_name,
        }
        return self.get(URL.COMMUNITY / f"inventory/{user_id64}/{app_id}/{context_id}", params=params)

    async def get_client_user_inventory(
        self, app_id: int, context_id: int, language: Language | None
    ) -> trade.Inventory:
        async with self.personal_inventory_lock:  # requires a lock
            return await self.get_user_inventory(self.user.id64, app_id, context_id, language)

    async def get_item_info(
        self, app_id: int, items: Iterable[tuple[int, int]], language: Language | None
    ) -> dict[tuple[int, int], trade.Description]:
        result: dict[tuple[int, int], trade.Description] = {}

        for chunk in utils.chunk(items, 100):
            params = {
                "key": self.api_key,
                "appid": app_id,
                "class_count": len(chunk),
                "language": (language or self.language).web_api_name,
            }

            for i, (class_id, instance_id) in enumerate(chunk):
                params[f"classid{i}"] = class_id
                params[f"instanceid{i}"] = instance_id

            data = await self.get(api_route("ISteamEconomy/GetAssetClassInfo"), params=params)
            result.update({tuple(map(int, key.split("_"))): value for key, value in data["result"].items()})

        return result

    def get_user_escrow(self, user_id64: int, token: str | None) -> ResponseType[dict[str, Any]]:
        params = {
            "key": self.api_key,
            "steamid_target": user_id64,
            "trade_offer_access_token": token if token is not None else "",
        }
        return self.get(api_route("IEconService/GetTradeHoldDurations"), params=params)

    async def get_friends(self, user_id64: int) -> list[User]:
        params = {"key": self.api_key, "steamid": user_id64, "relationship": "friend"}
        friends = await self.get(api_route("ISteamUser/GetFriendList"), params=params)
        return await self.get_users([friend["steamid"] for friend in friends["friendslist"]["friends"]])

    async def get_trade_offers(
        self,
        active_only: bool = True,
        sent: bool = True,
        received: bool = True,
        updated_only: bool = True,
        language: Language | None = None,
    ) -> dict[str, Any]:  # TODO consider making async iter?
        params = {
            "key": self.api_key,
            "active_only": str(active_only).lower(),
            "get_sent_offers": str(sent).lower(),
            "get_received_offers": str(received).lower(),
            "get_descriptions": "true",
            "cursor": 0,
            "language": (language or self.language).api_name,
        }
        if updated_only:
            try:
                params["time_historical_cutoff"] = self.trades_last_fetched
            except AttributeError:
                pass
        resp = await self.get(api_route("IEconService/GetTradeOffers"), params=params)
        first_page = resp["response"]
        next_cursor = first_page.get("next_cursor", 0)
        current_cursor = 0
        while current_cursor < next_cursor:
            params["cursor"] = next_cursor
            resp = await self.get(api_route("IEconService/GetTradeOffers"), params=params)
            page = resp["response"]
            for key, value in page.items():
                value_in_first_page = first_page[key]
                if isinstance(value_in_first_page, list):
                    value_in_first_page += value

            current_cursor = next_cursor
            try:
                next_cursor = page["next_cursor"]
            except KeyError:
                break

        if updated_only:
            self.trades_last_fetched = int(time())

        return first_page

    def get_trade_history(
        self, limit: int, previous_time: int = 0, language: Language | None = None
    ) -> ResponseType[trade.GetTradeOfferHistory]:
        params = {
            "key": self.api_key,
            "max_trades": limit,
            "get_descriptions": "true",
            "include_total": "true",
            "start_after_time": previous_time,
            "language": (language or self.language).api_name,
        }
        return self.get(api_route("IEconService/GetTradeHistory"), params=params)

    def get_trade(self, trade_id: int, language: Language | None = None) -> ResponseType[trade.GetTradeOffer]:
        params = {"key": self.api_key, "tradeofferid": trade_id, "get_descriptions": "true"}
        return self.get(api_route("IEconService/GetTradeOffer"), params=params)

    def accept_user_trade(self, user_id64: int, trade_id: int) -> Coro[dict[str, Any]]:
        payload = {
            "sessionid": self.session_id,
            "tradeofferid": trade_id,
            "serverid": 1,
            "partner": user_id64,
            "captcha": "",
        }
        headers = {"Referer": str(URL.COMMUNITY / f"tradeoffer/{trade_id}")}
        return self.post(URL.COMMUNITY / f"tradeoffer/{trade_id}/accept", data=payload, headers=headers)

    def _cancel_user_trade(self, trade_id: int, option: str) -> Coro[None]:
        payload = {"sessionid": self.session_id}
        return self.post(URL.COMMUNITY / f"tradeoffer/{trade_id}/{option}", data=payload)

    def decline_user_trade(self, trade_id: int) -> Coro[None]:
        return self._cancel_user_trade(trade_id, "decline")

    def cancel_user_trade(self, trade_id: int) -> Coro[None]:
        return self._cancel_user_trade(trade_id, "cancel")

    def send_trade_offer(
        self,
        user: SteamID,
        to_send: list[trade.AssetToDict],
        to_receive: list[trade.AssetToDict],
        token: str | None,
        offer_message: str,
        **kwargs: Any,
    ) -> Coro[trade.TradeOfferCreateResponse]:
        payload = {
            "sessionid": self.session_id,
            "serverid": 1,
            "partner": user.id64,
            "tradeoffermessage": offer_message,
            "json_tradeoffer": JSON_DUMPS(
                {
                    "newversion": True,
                    "version": len(to_send) + len(to_receive) + 1,
                    "me": {"assets": to_send, "currency": [], "ready": False},
                    "them": {"assets": to_receive, "currency": [], "ready": False},
                }
            ),
            "captcha": "",
            "trade_offer_create_params": JSON_DUMPS({"trade_offer_access_token": token}) if token is not None else "{}",
            **kwargs,
        }
        referer = URL.COMMUNITY / f"tradeoffer/new/?partner={user.id}"
        if token is not None:
            referer %= {"token": token}
        headers = {"Referer": str(referer)}
        return self.post(URL.COMMUNITY / "tradeoffer/new/send", data=payload, headers=headers)

    def get_trade_receipt(self, trade_id: int, language: Language | None = None) -> Coro[dict[str, Any]]:
        params = {
            "key": self.api_key,
            "tradeid": trade_id,
            "get_descriptions": "true",
            "language": (language or self.language).api_name,
        }
        return self.get(api_route("IEconService/GetTradeStatus"), params=params)

    def get_cm_list(self, cell_id: int) -> Coro[dict[str, Any]]:
        params = {
            "cellid": cell_id,
            "cmtype": "websockets",
        }
        return self.get(api_route("ISteamDirectory/GetCMListForConnect"), params=params)

    def join_clan(self, clan_id64: int) -> Coro[None]:
        payload = {
            "sessionID": self.session_id,
            "action": "join",
        }
        return self.post(URL.COMMUNITY / f"gid/{clan_id64}", data=payload)

    def leave_clan(self, clan_id64: int) -> Coro[None]:
        payload = {
            "sessionID": self.session_id,
            "action": "leaveGroup",
            "groupId": clan_id64,
        }
        return self.post(URL.COMMUNITY / "my/home_process", data=payload)

    def invite_user_to_clan(self, user_id64: int, clan_id64: int) -> Coro[None]:
        payload = {
            "sessionID": self.session_id,
            "group": clan_id64,
            "invitee": user_id64,
            "type": "groupInvite",
        }
        return self.post(URL.COMMUNITY / "actions/GroupInvite", data=payload)

    def get_user_clans(self, user_id64: int) -> Coro[dict[str, Any]]:
        params = {"key": self.api_key, "steamid": user_id64}
        return self.get(api_route("ISteamUser/GetUserGroupList"), params=params)

    def get_user_bans(self, user_id64: int) -> Coro[dict[str, Any]]:
        params = {"key": self.api_key, "steamids": user_id64}
        return self.get(api_route("ISteamUser/GetPlayerBans"), params=params)

    def get_user_level(self, user_id64: int) -> Coro[dict[str, Any]]:
        params = {"key": self.api_key, "steamid": user_id64}
        return self.get(api_route("IPlayerService/GetSteamLevel"), params=params)

    def get_user_badges(self, user_id64: int) -> Coro[dict[str, Any]]:
        params = {"key": self.api_key, "steamid": user_id64}
        return self.get(api_route("IPlayerService/GetBadges"), params=params)

    def clear_nickname_history(self) -> Coro[None]:
        payload = {"sessionid": self.session_id}
        return self.post(URL.COMMUNITY / "my/ajaxclearaliashistory", data=payload)

    def get_price(self, app_id: int, item_name: str, currency: int) -> Coro[PriceOverviewDict]:
        payload = {
            "appid": app_id,
            "market_hash_name": item_name,
        }
        payload.update({"currency": currency} if currency is not None else {})
        return self.post(URL.COMMUNITY / "market/priceoverview", data=payload)

    def get_wishlist(self, user_id64: int) -> Coro[dict[str, Any]]:
        return self.get(URL.STORE / f"wishlist/profiles/{user_id64}/wishlistdata")

    def get_game(self, game_id: int, language: Language | None) -> Coro[dict[str, Any]]:
        params = {
            "appids": game_id,
            "l": (language or self.language).api_name,
        }
        return self.get(URL.STORE / "api/appdetails", params=params)

    def get_game_dlc(self, game_id: int, language: Language | None) -> Coro[dict[str, Any]]:
        params = {
            "appid": game_id,
            "l": (language or self.language).api_name,
        }
        return self.get(URL.STORE / "api/dlcforapp", params=params)

    def get_clan_rss(self, clan_id64: int) -> Coro[str]:
        return self.get(URL.COMMUNITY / f"gid/{clan_id64}/rss")

    def _edit_clan_event(
        self,
        action: str,
        clan_id64: int,
        name: str,
        description: str,
        event_type: str,
        game_id: str,
        server_ip: str,
        server_password: str,
        start: datetime | None,
        event_id: int | None,
    ) -> Coro[str]:
        if start is None:
            tz_offset = int((datetime.now().astimezone().tzinfo.utcoffset()).total_seconds())
            start_date = "MM/DD/YY"
            start_hour = "12"
            start_minute = "00"
            start_ampm = "PM"
            time_choice = "quick"
        else:
            if start.tzinfo is None:
                start = start.astimezone()

            offset = start.tzinfo.utcoffset(start)
            tz_offset = int(offset.total_seconds())

            start_date = f"{start:%m/%d/%y}"
            start_hour = f"{start:%I}"
            start_minute = f"{start:%M}"
            start_ampm = "AM" if start.hour <= 12 else "PM"
            time_choice = "specific"

        data = {
            "sessionid": self.session_id,
            "action": action,
            "tzOffset": tz_offset,
            "name": name,
            "type": event_type,
            "appID": game_id,
            "serverIP": server_ip,
            "serverPassword": server_password,
            "notes": description,
            "eventQuickTime": "now",
            "startDate": start_date,
            "startHour": start_hour,
            "startMinute": start_minute,
            "startAMPM": start_ampm,
            "timeChoice": time_choice,
        }

        if event_id is not None:
            data["eventID"] = event_id

        return self.post(URL.COMMUNITY / f"gid/{clan_id64}/eventEdit", data=data)

    def create_clan_event(self, *args: Any, **kwargs: Any) -> Coro[str]:
        return self._edit_clan_event("newEvent", *args, event_id=None, **kwargs)

    def edit_clan_event(self, *args: Any, **kwargs: Any) -> Coro[str]:
        return self._edit_clan_event("updateEvent", *args, **kwargs)

    def delete_clan_event(self, clan_id64: int, event_id: int) -> Coro[None]:
        data = {
            "sessionid": self.session_id,
            "action": "deleteEvent",
            "eventID": event_id,
        }
        return self.post(URL.COMMUNITY / f"gid/{clan_id64}/events", data=data)

    def get_clan_events(self, clan_id: int, event_ids: list[int]) -> Coro[dict[str, Any]]:
        params = {
            "clanid_list": ",".join([str(clan_id)] * len(event_ids)),
            "uniqueid_list": ",".join(str(id) for id in event_ids),
        }
        return self.get(URL.STORE / "events/ajaxgeteventdetails", params=params)

    def create_clan_announcement(self, clan_id64: int, name: str, description: str, hidden: bool = False) -> Coro[None]:
        data = {
            "sessionID": self.session_id,
            "action": "post",
            "headline": name,
            "body": description,
            "languages[0][headline]": name,
            "languages[0][body]": description,
        }
        if hidden:
            data["is_hidden"] = "is_hidden"
        return self.post(URL.COMMUNITY / f"gid/{clan_id64}/announcements", data=data)

    def edit_clan_announcement(self, clan_id64: int, announcement_id: int, name: str, description: str) -> Coro[None]:
        data = {
            "sessionID": self.session_id,
            "gid": announcement_id,
            "action": "update",
            "headline": name,
            "body": description,
            "languages[0][headline]": name,
            "languages[0][body]": description,
            "languages[0][updated]": 1,
        }
        return self.post(URL.COMMUNITY / f"gid/{clan_id64}/announcements", data=data)

    def delete_clan_announcement(self, clan_id64: int, announcement_id: int) -> Coro[None]:
        params = {
            "sessionID": self.session_id,
        }
        return self.post(URL.COMMUNITY / f"gid/{clan_id64}/announcements/delete/{announcement_id}", params=params)

    def get_clan_announcement(
        self,
        clan_id: int,
        announcement_id: int,
    ) -> Coro[dict[str, Any]]:
        params = {
            "clan_accountid": clan_id,
            "announcement_gid": announcement_id,
        }
        return self.get(URL.STORE / "events/ajaxgetpartnerevent", params=params)

    def post_review(
        self,
        game_id: int,
        content: str,
        upvoted: bool,
        public: bool,
        commentable: bool,
        received_compensation: bool,
        language: str,
    ) -> Coro[None]:
        data = {
            "appid": game_id,
            "steamworksappid": game_id,
            "comment": content,
            "rated_up": str(upvoted).lower(),
            "is_public": str(public).lower(),
            "language": language,
            "received_compensation": int(received_compensation),
            "disable_comments": int(not commentable),
            "sessionid": self.session_id,
        }

        return self.post(URL.STORE / "friends/recommendgame", data=data)

    def get_reviews(
        self, game_id: int, filter: str, review_type: str, purchase_type: str, cursor: str = "*"
    ) -> Coro[dict[str, Any]]:
        params = {
            "json": 1,
            "num_per_page": 100,
            "cursor": urllib.parse.quote(cursor),
            "filter": filter,
            "review_type": review_type,
            "purchase_type": purchase_type,
        }

        return self.get(URL.STORE / f"appreviews/{game_id}", params=params)

    def mark_review_as_helpful(self, review_id: int, rated_up: bool) -> Coro[None]:
        data = {
            "rateup": str(rated_up).lower(),
            "sessionid": self.session_id,
        }
        return self.post(URL.COMMUNITY / f"userreviews/rate/{review_id}", data=data)

    def mark_review_as_funny(self, review_id: int) -> Coro[None]:
        data = {
            "tagid": 1,
            "rateup": "true",
            "sessionid": self.session_id,
        }
        return self.post(URL.COMMUNITY / f"userreviews/votetag/{review_id}", data=data)

    def delete_review(self, game_id: int) -> Coro[None]:
        data = {
            "action": "delete",
            "appid": game_id,
            "sessionid": self.session_id,
        }
        return self.post(URL.COMMUNITY / "my/recommended", data=data)

    def get_package(self, package_id: int, language: Language | None) -> Coro[dict[str, Any]]:
        params = {
            "packageids": package_id,
            "l": (language or self.language).api_name,
        }
        return self.get(URL.STORE / "api/packagedetails", params=params)

    async def edit_profile_info(
        self,
        name: str | None,
        real_name: str | None,
        url: str | None,
        summary: str | None,
        country: str | None,
        state: str | None,
        city: str | None,
    ) -> None:
        info = await self.user.profile_info()

        payload = {
            "sessionID": self.session_id,
            "type": "profileSave",
            "weblink_1_title": "",
            "weblink_1_url": "",
            "weblink_2_title": "",
            "weblink_2_url": "",
            "weblink_3_title": "",
            "weblink_3_url": "",
            "personaName": name or self.user.name,
            "real_name": real_name or info.real_name,
            "customURL": url or self.user.community_url,
            "country": country or info.country_name,
            "state": state or info.state_name,
            "city": city or info.city_name,
            "summary": summary or info.summary,
        }

        await self.post(URL.COMMUNITY / "my/edit", data=payload)

    async def update_avatar(self, avatar: Image) -> None:
        with avatar:
            payload = aiohttp.FormData()
            payload.add_field("MAX_FILE_SIZE", str(avatar.size))
            payload.add_field("type", "player_avatar_image")
            payload.add_field("sId", str(self.user.id64))
            payload.add_field("sessionid", self.session_id)
            payload.add_field("doSub", "1")
            payload.add_field(
                "avatar", avatar.read(), filename=f"avatar.{avatar.type}", content_type=f"image/{avatar.type}"
            )
            await self.post(URL.COMMUNITY / "actions/FileUploader", data=payload)

    async def send_user_image(self, user_id64: int, image: Image) -> None:
        with image:
            payload = {
                "sessionid": self.session_id,
                "l": "english",
                "file_size": image.size,
                "file_name": image.name,
                "file_sha": image.hash,
                "file_image_width": image.width,
                "file_image_height": image.height,
                "file_type": f"image/{image.type}",
            }
            resp = await self.post(URL.COMMUNITY / "chat/beginfileupload", data=payload)

            result = resp["result"]
            url = f'{"https" if result["use_https"] else "http"}://{result["url_host"]}{result["url_path"]}'
            headers = {header["name"]: header["value"] for header in result["request_headers"]}
            await self.request("PUT", url=url, headers=headers, data=image.read())

            payload.update(
                {
                    "success": 1,
                    "ugcid": result["ugcid"],
                    "timestamp": resp["timestamp"],
                    "hmac": resp["hmac"],
                    "friend_steamid": user_id64,
                    "spoiler": int(image.spoiler),
                }
            )
            await self.post(URL.COMMUNITY / "chat/commitfileupload", data=payload)

    async def send_chat_image(self, chat_id: int, channel_id: int, image: Image) -> None:
        with image:
            payload = {
                "sessionid": self.session_id,
                "l": "english",
                "file_size": image.size,
                "file_name": image.name,
                "file_sha": image.hash,
                "file_image_width": image.width,
                "file_image_height": image.height,
                "file_type": f"image/{image.type}",
            }
            resp = await self.post(URL.COMMUNITY / "chat/beginfileupload", data=payload)

            result = resp["result"]
            url = f'{"https" if result["use_https"] else "http"}://{result["url_host"]}{result["url_path"]}'
            headers = {header["name"]: header["value"] for header in result["request_headers"]}
            await self.request("PUT", url=url, headers=headers, data=image.read())

            payload.update(
                {
                    "success": 1,
                    "ugcid": result["ugcid"],
                    "timestamp": resp["timestamp"],
                    "hmac": resp["hmac"],
                    "chat_group_id": channel_id,
                    "chat_id": chat_id,
                    "spoiler": int(image.spoiler),
                }
            )
            await self.post(URL.COMMUNITY / "chat/commitfileupload", data=payload)

    async def get_api_key(self) -> str | None:
        resp = await self.get(URL.COMMUNITY / "dev/apikey")
        if (
            "<h2>Access Denied</h2>" in resp
            or "You must have a validated email address to create a Steam Web API key" in resp
        ):
            return

        key_re = re.compile(r"<p>Key: ([0-9A-F]+)</p>")
        match = key_re.findall(resp)
        if match:
            return match[0]

        payload = {
            "domain": "steam.py",
            "agreeToTerms": "agreed",
            "sessionid": self.session_id,
            "Submit": "Register",
        }
        resp = await self.post(URL.COMMUNITY / "dev/registerkey", data=payload)
        return key_re.findall(resp)[0]



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ...errors import SteamException

if TYPE_CHECKING:
    from inspect import Parameter

    from .commands import Command


__all__ = (
    "CommandError",
    "BadArgument",
    "MissingRequiredArgument",
    "DuplicateKeywordArgument",
    "UnmatchedKeyValuePair",
    "CheckFailure",
    "NotOwner",
    "CommandNotFound",
    "CommandDisabled",
    "CommandOnCooldown",
)


class CommandError(SteamException):
    """Base Exception for errors raised by commands.

    Subclass of :exc:`SteamException`.
    """


class CommandNotFound(CommandError):
    """Exception raised when a command is not found.

    Subclass of :exc:`CommandError`.
    """


class BadArgument(CommandError):
    """Exception raised when a bad argument is passed to a command.

    Subclass of :exc:`CommandError`.
    """


class MissingRequiredArgument(BadArgument):
    """Exception raised when a required argument is not passed to a command.

    Subclass of :exc:`BadArgument`.

    Attributes
    ----------
    param
        The argument that is missing.
    """

    def __init__(self, param: Parameter):
        self.param = param
        super().__init__(f"{param.name!r} is a required argument that is missing.")


class DuplicateKeywordArgument(BadArgument):
    """Exception raised when a keyword argument passed would shadow another.

    Subclass of :exc:`BadArgument`.

    Attributes
    ----------
    name
        The argument that would shadow another.
    """

    def __init__(self, name: str):
        self.name = name
        super().__init__(f"{name!r} shadows another argument.")


class UnmatchedKeyValuePair(BadArgument):
    """Exception raised when an incorrect number of key value pairs are passed to a command to be unpacked.

    Subclass of :exc:`BadArgument`.
    """


class CheckFailure(CommandError):
    """Base Exception raised when a check fails.

    Subclass of :exc:`CommandError`.
    """


class CommandDisabled(CheckFailure):
    """Exception raised when a command is disabled and is attempted to be ran.

    Subclass of :exc:`CheckFailure`.

    Attributes
    -----------
    command
        The command that has been disabled.
    """

    def __init__(self, command: Command):
        self.command = command
        super().__init__(f"{command.name} is currently disabled")


class NotOwner(CheckFailure):
    """Exception raised the user does not own the bot.

    Subclass of :exc:`CheckFailure`.
    """

    def __init__(self):
        super().__init__("You do not own this bot")


class DMChannelOnly(CheckFailure):
    """Exception raised the user does not own the bot.

    Subclass of :exc:`CheckFailure`.
    """

    def __init__(self):
        super().__init__("This command can only be used in DMs")


class CommandOnCooldown(CommandError):
    """Exception raised when a command is still on cooldown.

    Subclass of :exc:`CommandError`.

    Attributes
    ----------
    retry_after
        The time in seconds at which that the next command can successfully be executed.
    """

    def __init__(self, retry_after: float):
        self.retry_after = retry_after
        super().__init__(f"Command is on cooldown for {retry_after:.2} more seconds")


class MissingClosingQuotation(CommandError):
    def __init__(self, position: int):
        self.position = position
        super().__init__(f"No closing quotation found after the character at position {position}")



"""
Various constants/types for use around the library.

Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE
"""

from __future__ import annotations

import builtins
import sys
from collections.abc import Callable
from io import StringIO
from typing import Any

from multidict import MultiDict
from typing_extensions import Final, final
from yarl import URL as _URL

from .types.vdf import BinaryVDFDict, VDFDict

DOCS_BUILDING: bool = getattr(builtins, "__sphinx__", False)
TASK_HAS_NAME = sys.version_info >= (3, 8)


try:
    import orjson
except ModuleNotFoundError:
    import json

    JSON_LOADS: Final = json.loads

    def dumps(
        obj: Any,
        *,
        __func: Callable[..., str] = json.dumps,
    ) -> str:
        return __func(obj, separators=(",", ":"), ensure_ascii=True)

    JSON_DUMPS: Final = dumps
else:
    JSON_LOADS: Final[Callable[[str | bytes], Any]] = orjson.loads

    def dumps(
        obj: Any,
        *,
        __func: Callable[[Any], bytes] = orjson.dumps,
        __decoder: Callable[[bytes], str] = bytes.decode,
    ) -> str:
        return __decoder(__func(obj))

    JSON_DUMPS: Final = dumps


try:
    import orvdf
except ModuleNotFoundError:
    import vdf

    def loads(
        s: str,
        *,
        __func: Callable[..., Any] = vdf.parse,
        __mapper: type[MultiDict[Any]] = MultiDict,
        __string_io: type[StringIO] = StringIO,
    ) -> VDFDict:
        return __func(__string_io(s), mapper=__mapper)

    def binary_loads(
        s: bytes, *, __func: Callable[..., Any] = vdf.binary_loads, __mapper: type[MultiDict[Any]] = MultiDict
    ) -> BinaryVDFDict:
        return __func(s, mapper=__mapper)

    VDF_LOADS: Final = loads
    VDF_BINARY_LOADS: Final = binary_loads

else:
    VDF_LOADS: Final[Callable[[str], VDFDict]] = orvdf.loads
    VDF_BINARY_LOADS: Final[Callable[[bytes], BinaryVDFDict]] = orvdf.binary_loads


try:
    import lxml
except ModuleNotFoundError:
    HTML_PARSER: Final = "html.parser"
else:
    HTML_PARSER: Final = "lxml-xml"


@final
class URL:
    API: Final = _URL("https://api.steampowered.com")
    COMMUNITY: Final = _URL("https://steamcommunity.com")
    STORE: Final = _URL("https://store.steampowered.com")


# default CMs if Steam API is down
DEFAULT_CMS: Final = (
    "cm1-lhr1.cm.steampowered.com:27019",
    "cm1-lhr1.cm.steampowered.com:27020",
    "cm1-lhr1.cm.steampowered.com:27021",
    "cm1-lhr1.cm.steampowered.com:27022",
    "cm1-lhr1.cm.steampowered.com:27023",
    "cm1-lhr1.cm.steampowered.com:27024",
    "cm1-lhr1.cm.steampowered.com:27025",
    "cm1-lhr1.cm.steampowered.com:27028",
    "cm1-lhr1.cm.steampowered.com:27029",
    "cm1-lhr1.cm.steampowered.com:27030",
    "cm1-lhr1.cm.steampowered.com:27031",
    "cm1-lhr1.cm.steampowered.com:27032",
    "cm1-lhr1.cm.steampowered.com:27033",
    "cm1-lhr1.cm.steampowered.com:27034",
    "cm1-lhr1.cm.steampowered.com:27035",
    "cm1-lhr1.cm.steampowered.com:27036",
    "cm1-lhr1.cm.steampowered.com:27037",
    "cm1-lhr1.cm.steampowered.com:27038",
    "cm1-lhr1.cm.steampowered.com:443",
    "cm2-lhr1.cm.steampowered.com:27019",
    "cm2-lhr1.cm.steampowered.com:27020",
    "cm2-lhr1.cm.steampowered.com:27021",
    "cm2-lhr1.cm.steampowered.com:27022",
    "cm2-lhr1.cm.steampowered.com:27023",
    "cm2-lhr1.cm.steampowered.com:27024",
    "cm2-lhr1.cm.steampowered.com:27025",
    "cm2-lhr1.cm.steampowered.com:27028",
    "cm2-lhr1.cm.steampowered.com:27029",
    "cm2-lhr1.cm.steampowered.com:27030",
    "cm2-lhr1.cm.steampowered.com:27031",
    "cm2-lhr1.cm.steampowered.com:27032",
    "cm2-lhr1.cm.steampowered.com:27033",
    "cm2-lhr1.cm.steampowered.com:27034",
    "cm2-lhr1.cm.steampowered.com:27035",
    "cm2-lhr1.cm.steampowered.com:27036",
    "cm2-lhr1.cm.steampowered.com:27037",
    "cm2-lhr1.cm.steampowered.com:27038",
    "cm2-lhr1.cm.steampowered.com:443",
    "cm3-lhr1.cm.steampowered.com:27019",
    "cm3-lhr1.cm.steampowered.com:27020",
    "cm3-lhr1.cm.steampowered.com:27021",
    "cm3-lhr1.cm.steampowered.com:27022",
    "cm3-lhr1.cm.steampowered.com:27023",
    "cm3-lhr1.cm.steampowered.com:27024",
    "cm3-lhr1.cm.steampowered.com:27025",
    "cm3-lhr1.cm.steampowered.com:27028",
    "cm3-lhr1.cm.steampowered.com:27029",
    "cm3-lhr1.cm.steampowered.com:27030",
    "cm3-lhr1.cm.steampowered.com:27031",
    "cm3-lhr1.cm.steampowered.com:27032",
    "cm3-lhr1.cm.steampowered.com:27033",
    "cm3-lhr1.cm.steampowered.com:27034",
    "cm3-lhr1.cm.steampowered.com:27035",
    "cm3-lhr1.cm.steampowered.com:27036",
    "cm3-lhr1.cm.steampowered.com:27037",
    "cm3-lhr1.cm.steampowered.com:27038",
    "cm3-lhr1.cm.steampowered.com:443",
    "cm4-lhr1.cm.steampowered.com:27019",
    "cm4-lhr1.cm.steampowered.com:27020",
    "cm4-lhr1.cm.steampowered.com:27021",
    "cm4-lhr1.cm.steampowered.com:27022",
    "cm4-lhr1.cm.steampowered.com:27023",
    "cm4-lhr1.cm.steampowered.com:27024",
    "cm4-lhr1.cm.steampowered.com:27025",
    "cm4-lhr1.cm.steampowered.com:27028",
    "cm4-lhr1.cm.steampowered.com:27029",
    "cm4-lhr1.cm.steampowered.com:27030",
    "cm4-lhr1.cm.steampowered.com:27031",
    "cm4-lhr1.cm.steampowered.com:27032",
    "cm4-lhr1.cm.steampowered.com:27033",
    "cm4-lhr1.cm.steampowered.com:27034",
    "cm4-lhr1.cm.steampowered.com:27035",
    "cm4-lhr1.cm.steampowered.com:27036",
    "cm4-lhr1.cm.steampowered.com:27037",
    "cm4-lhr1.cm.steampowered.com:27038",
    "cm4-lhr1.cm.steampowered.com:443",
    "cm6-ams1.cm.steampowered.com:27023",
    "cm6-ams1.cm.steampowered.com:27025",
    "cm6-ams1.cm.steampowered.com:27032",
    "cm6-ams1.cm.steampowered.com:27033",
)



"""
Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE

Contains large portions of:
https://github.com/ValvePython/steam/blob/master/steam/steamid.py
The appropriate license is in LICENSE
"""

from __future__ import annotations

import abc
import asyncio
import re
from collections.abc import Coroutine
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any, TypeVar

import attr
from typing_extensions import ClassVar, Final, Protocol, Required, Self, TypedDict, runtime_checkable

from ._const import URL
from .badge import FavouriteBadge, UserBadges
from .enums import *
from .errors import WSException
from .game import Game, StatefulGame, UserGame, WishlistGame
from .iterators import AsyncIterator, CommentsIterator, UserPublishedFilesIterator, UserReviewsIterator
from .models import Ban
from .profile import *
from .reaction import Award, AwardReaction, Emoticon, MessageReaction, PartialMessageReaction, Sticker
from .trade import Inventory
from .utils import (
    _INVITE_HEX,
    _INVITE_MAPPING,
    DateTime,
    InstanceType,
    Intable,
    TypeType,
    UniverseType,
    cached_slot_property,
    id64_from_url,
    make_id64,
)

if TYPE_CHECKING:
    from aiohttp import ClientSession

    from .clan import Clan
    from .comment import Comment
    from .group import Group
    from .http import StrOrURL
    from .image import Image
    from .message import Authors
    from .protobufs.chat import Mentions
    from .review import Review
    from .state import ConnectionState
    from .types.id import ID32, ID64
    from .user import User

__all__ = (
    "SteamID",
    "Message",
    "Channel",
)

C = TypeVar("C", bound="Commentable")
M_co = TypeVar("M_co", bound="Message", covariant=True)


# TODO when defaults are implemented, make this Generic over Literal[Type] maybe
# TypeT = typing.TypeVar("TypeT", bound=Type)
#
#
# class SteamID(typing.Generic[TypeT]):
#     type TypeT
class SteamID(metaclass=abc.ABCMeta):
    """Convert a Steam ID between its various representations.

    Note
    ----
    See :func:`steam.utils.make_id64` for the full parameter list.
    """

    __slots__ = ("__BASE",)

    def __init__(
        self,
        id: Intable = 0,
        type: TypeType | None = None,
        universe: UniverseType | None = None,
        instance: InstanceType | None = None,
    ):
        self.__BASE: Final = make_id64(id, type, universe, instance)

    def __int__(self) -> ID64:
        return self.__BASE

    def __eq__(self, other: Any) -> bool:
        try:
            return self.__BASE == int(other)
        except (TypeError, ValueError):
            return NotImplemented

    def __str__(self) -> str:
        return str(self.__BASE)

    def __hash__(self) -> int:
        return hash(self.__BASE)

    def __repr__(self) -> str:
        return f"SteamID(id={self.id}, type={self.type}, universe={self.universe}, instance={self.instance})"

    @property
    def instance(self) -> InstanceFlag:
        """The instance of the SteamID."""
        return InstanceFlag.try_value((self.__BASE >> 32) & 0xFFFFF)

    @property
    def type(self) -> Type:
        """The Steam type of the SteamID."""
        return Type((self.__BASE >> 52) & 0xF)

    @property
    def universe(self) -> Universe:
        """The Steam universe of the SteamID."""
        return Universe((self.__BASE >> 56) & 0xFF)

    @property
    def id64(self) -> ID64:
        """The SteamID's 64-bit ID."""
        return self.__BASE

    @property
    def id(self) -> ID32:
        """The SteamID's 32-bit ID."""
        return self.__BASE & 0xFFFFFFFF

    @property
    def id2(self) -> str:
        """The SteamID's ID 2.

        e.g ``STEAM_1:0:1234``.
        """
        return f"STEAM_{self.universe.value}:{self.id % 2}:{self.id >> 1}"

    @property
    def id2_zero(self) -> str:
        """The SteamID's ID 2 accounted for bugged GoldSrc and Orange Box games.

        Note
        ----
        In these games the accounts :attr:`universe`, ``1`` for :class:`.Type.Public`, should be the ``X`` component of
        ``STEAM_X:0:1234`` however, this was bugged and the value of ``X`` was ``0``.

        e.g ``STEAM_0:0:1234``.
        """
        return self.id2.replace("_1", "_0")

    @property
    def id3(self) -> str:
        """The SteamID's ID 3.

        e.g ``[U:1:1234]``.
        """
        type_char = TypeChar(self.type).name
        instance = None

        if self.type in (Type.AnonGameServer, Type.Multiseat):
            instance = self.instance
        elif self.type == Type.Individual:
            if self.instance != InstanceFlag.Desktop:
                instance = self.instance
        elif self.type == Type.Chat:
            if self.instance & InstanceFlag.ChatClan > 0:
                type_char = "c"
            elif self.instance & InstanceFlag.ChatLobby > 0:
                type_char = "L"
            else:
                type_char = "T"

        return f"[{type_char}:{self.universe.value}:{self.id}{f':{instance.value}' if instance is not None else ''}]"

    @property
    def invite_code(self) -> str | None:
        """The SteamID's invite code in the s.team invite code format.

        e.g. ``cv-dgb``.
        """
        if self.type == Type.Individual and self.is_valid():
            invite_code = re.sub(f"[{_INVITE_HEX}]", lambda x: _INVITE_MAPPING[x.group()], f"{self.id:x}")
            split_idx = len(invite_code) // 2
            return invite_code if split_idx == 0 else f"{invite_code[:split_idx]}-{invite_code[split_idx:]}"

    @property
    def invite_url(self) -> str | None:
        """The SteamID's full invite code URL.

        e.g ``https://s.team/p/cv-dgb``.
        """
        code = self.invite_code
        return f"https://s.team/p/{code}" if code else None

    @property
    def community_url(self) -> str | None:
        """The SteamID's community url.

        e.g https://steamcommunity.com/profiles/123456789.
        """
        suffix = {
            Type.Individual: "profiles",
            Type.Clan: "gid",
        }
        try:
            return f"https://steamcommunity.com/{suffix[self.type]}/{self.id64}"
        except KeyError:
            return None

    def is_valid(self) -> bool:
        """Whether the SteamID is valid."""
        if self.type == Type.Invalid or self.type >= Type.Max:
            return False

        if self.universe == Universe.Invalid or self.universe >= Universe.Max:
            return False

        if self.type == Type.Individual and (self.id == 0 or self.instance > 4):
            return False

        if self.type == Type.Clan and (self.id == 0 or self.instance != 0):
            return False

        if self.type == Type.GameServer and self.id == 0:
            return False

        if self.type == Type.AnonGameServer and self.id == 0 and self.instance == 0:
            return False

        return True

    @staticmethod
    async def from_url(url: StrOrURL, session: ClientSession | None = None) -> SteamID | None:
        """A helper function creates a SteamID instance from a Steam community url.

        Note
        ----
        See :func:`id64_from_url` for the full parameter list.
        """
        id64 = await id64_from_url(url, session)
        return SteamID(id64) if id64 else None


class _CommentableKwargs(TypedDict, total=False):
    id64: Required[int]
    thread_type: Required[int]
    gidfeature: int
    gidfeature2: int
    thread_id: int
    upvotes: int
    include_deleted: bool


class Commentable(Protocol):
    """A mixin that implements commenting functionality."""

    __slots__ = ()
    _state: ConnectionState

    @property
    @abc.abstractmethod
    def _commentable_kwargs(self) -> _CommentableKwargs:
        raise NotImplementedError

    async def fetch_comment(self, id: int) -> Comment[Self]:
        """Fetch a comment by its ID.

        Parameters
        ----------
        id
            The ID of the comment to fetch.
        """
        from .comment import Comment

        comment = await self._state.fetch_comment(self, id)
        return Comment(
            self._state,
            id=comment.id,
            content=comment.content,
            created_at=DateTime.from_timestamp(comment.timestamp),
            author=await self._state._maybe_user(comment.author_id64),
            owner=self,
            reactions=[AwardReaction(self._state, reaction) for reaction in comment.reactions],
        )

    async def comment(self, content: str, *, subscribe: bool = True) -> Comment[Self]:
        """Post a comment to a comments section.

        Parameters
        ----------
        content
            The message to add to the comment section.
        subscribe
            Whether to subscribe to notifications on any future activity in this comment's thread.

        Returns
        -------
        The created comment.
        """
        return await self._state.post_comment(self, content, subscribe)

    def comments(
        self,
        *,
        oldest_first: bool = False,
        limit: int | None = None,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> CommentsIterator[Self]:
        """An :class:`~steam.iterators.AsyncIterator` for accessing a comment section's :class:`~steam.Comment` objects.

        Examples
        --------

        Usage:

        .. code-block:: python3

            async for comment in commentable.comments(limit=10):
                print("Author:", comment.author, "Said:", comment.content)

        Flattening into a list:

        .. code-block:: python3

            comments = await commentable.comments(limit=50).flatten()
            # comments is now a list of Comment

        All parameters are optional.

        Parameters
        ----------
        oldest_first
            Whether or not to request comments with the oldest comments first or last. Defaults to ``False``.
        limit
            The maximum number of comments to search through.
            Default is ``None`` which will fetch the all the comments in the comments section.
        before
            A time to search for comments before.
        after
            A time to search for comments after.

        Yields
        ---------
        :class:`~steam.Comment`
        """
        return CommentsIterator(
            oldest_first=oldest_first, owner=self, state=self._state, limit=limit, before=before, after=after
        )


class Awardable(Protocol):
    """A mixin that implements award functionality."""

    __slots__ = ()

    id: int
    _state: ConnectionState
    _AWARDABLE_TYPE: ClassVar[int]

    async def award(self, award: Award) -> None:
        """Add an :class:`Award` to this piece of user generated content.

        Parameters
        ----------
        award
            The award to add.
        """
        await self._state.add_award(self, award)

    # async def fetch_reactions(self) -> list[AwardReaction]:
    #     """Fetch the reactions on this piece of user generated content."""
    #     reactions = await self._state.fetch_award_reactions(self)
    #     return [AwardReaction(self._state, reaction) for reaction in reactions]


class BaseUser(SteamID, Commentable):
    """An ABC that details the common operations on a Steam user.
    The following classes implement this ABC:

        - :class:`~steam.User`
        - :class:`~steam.ClientUser`

    .. container:: operations

        .. describe:: x == y

            Checks if two users are equal.

        .. describe:: str(x)

            Returns the user's name.

    Attributes
    ----------
    name
        The user's username.
    state
        The current persona state of the account (e.g. LookingToTrade).
    game
        The Game instance attached to the user. Is ``None`` if the user isn't in a game or one that is recognised by the
        api.
    primary_clan
        The primary clan the User displays on their profile.

        Note
        ----
        This can be lazily awaited to get more attributes of the clan.

    avatar_url
        The avatar url of the user. Uses the large (184x184 px) image url.
    real_name
        The user's real name defined by them. Could be ``None``.
    created_at
        The time at which the user's account was created. Could be ``None``.
    last_logoff
        The last time the user logged into steam. Could be None (e.g. if they are currently online).
    country
        The country code of the account. Could be ``None``.
    flags
        The persona state flags of the account.
    rich_presence
        The user's rich presence.
    """

    __slots__ = (
        "name",
        "game",
        "state",
        "flags",
        "country",
        "primary_clan",
        "trade_url",
        "real_name",
        "avatar_url",
        "last_seen_online",
        "created_at",
        "last_logoff",
        "last_logon",
        "rich_presence",
        "privacy_state",
        "community_url",
        "comment_permissions",
        "profile_state",
        "_level",
        "_state",
        "__weakref__",
    )

    name: str
    real_name: str | None
    community_url: str | None
    avatar_url: str | None  # TODO make this a property and add avatar hash
    primary_clan: Clan | None
    country: str | None
    created_at: datetime | None
    last_logoff: datetime | None
    last_logon: datetime | None
    last_seen_online: datetime | None
    game: StatefulGame | None
    state: PersonaState | None
    flags: PersonaStateFlag | None
    privacy_state: CommunityVisibilityState | None
    comment_permissions: CommentPrivacyState | None
    profile_state: CommunityVisibilityState | None
    rich_presence: dict[str, str] | None
    _state: ConnectionState

    def __repr__(self) -> str:
        attrs = ("name", "state", "id", "universe", "instance")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    def __str__(self) -> str:
        return self.name

    @property
    def _commentable_kwargs(self) -> _CommentableKwargs:
        return {
            "id64": self.id64,
            "thread_type": 10,
        }

    @property
    def mention(self) -> str:
        """The string used to mention the user in chat."""
        return f"[mention={self.id}]@{self.name}[/mention]"

    async def inventory(self, game: Game, *, language: Language | None = None) -> Inventory:
        """Fetch a user's :class:`~steam.Inventory` for trading.

        Parameters
        -----------
        game
            The game to fetch the inventory for.
        language
            The language to fetch the inventory in. If ``None`` will default to the current language.

        Raises
        ------
        :exc:`~steam.Forbidden`
            The user's inventory is private.
        """
        resp = await self._state.http.get_user_inventory(self.id64, game.id, game.context_id, language)
        return Inventory(state=self._state, data=resp, owner=self, game=game, language=language)

    async def friends(self) -> list[User]:
        """Fetch the list of the users friends."""
        friends = await self._state.http.get_friends(self.id64)
        return [self._state._store_user(friend) for friend in friends]

    async def games(self, *, include_free: bool = True) -> list[UserGame]:
        r"""Fetches the :class:`~steam.Game`\s the user owns.

        Parameters
        ----------
        include_free
            Whether to include free games in the list. Defaults to ``True``.
        """
        games = await self._state.fetch_user_games(self.id64, include_free)
        return [UserGame(self._state, game) for game in games]

    async def clans(self) -> list[Clan]:
        r"""Fetches a list of :class:`~steam.Clan`\s the user is in."""

        async def getter(gid: int) -> Clan:
            try:
                clan = await self._state.client.fetch_clan(gid)
                assert clan is not None
                return clan
            except WSException as exc:
                if exc.code == Result.RateLimitExceeded:
                    await asyncio.sleep(20)
                    return await getter(gid)
                raise

        resp = await self._state.http.get_user_clans(self.id64)
        return await asyncio.gather(*(getter(int(clan["gid"])) for clan in resp["response"]["groups"]))  # type: ignore

    async def bans(self) -> Ban:
        r"""Fetches the user's :class:`.Ban`\s."""
        resp = await self._state.http.get_user_bans(self.id64)
        resp = resp["players"][0]
        resp["EconomyBan"] = resp["EconomyBan"] != "none"
        return Ban(data=resp)

    async def badges(self) -> UserBadges:
        r"""Fetches the user's :class:`.UserBadges`\s."""
        resp = await self._state.http.get_user_badges(self.id64)
        return UserBadges(self._state, data=resp["response"])

    async def level(self) -> int:
        """Fetches the user's level if your account is premium, otherwise it's cached."""
        if self._state.http.api_key is not None:
            resp = await self._state.http.get_user_level(self.id64)
            return resp["response"]["player_level"]
        return self._level

    async def wishlist(self) -> list[WishlistGame]:
        r"""Get the :class:`.WishlistGame`\s the user has on their wishlist."""
        data = await self._state.http.get_wishlist(self.id64)
        return [WishlistGame(self._state, id=id, data=game_info) for id, game_info in data.items()]

    async def favourite_badge(self) -> FavouriteBadge | None:
        """The user's favourite badge."""
        badge = await self._state.fetch_user_favourite_badge(self.id64)
        if not badge.has_favorite_badge:
            return

        return FavouriteBadge(
            id=UserBadge.try_value(badge.badgeid),
            item_id=badge.communityitemid,
            type=badge.item_type,
            border_colour=badge.border_color,
            game=StatefulGame(self._state, id=badge.appid) if badge.appid else None,
            level=badge.level,
        )

    async def equipped_profile_items(self, *, language: Language | None = None) -> EquippedProfileItems:
        """The user's equipped profile items.

        Parameters
        ----------
        language
            The language to fetch the profile items in. If ``None`` the current language is used
        """
        items = await self._state.fetch_user_equipped_profile_items(self.id64, language)
        return EquippedProfileItems(
            background=ProfileItem(self._state, self, items.profile_background) if items.profile_background else None,
            mini_profile_background=(
                ProfileItem(self._state, self, items.mini_profile_background) if items.mini_profile_background else None
            ),
            avatar_frame=ProfileItem(self._state, self, items.avatar_frame) if items.avatar_frame else None,
            animated_avatar=ProfileItem(self._state, self, items.animated_avatar) if items.animated_avatar else None,
            modifier=ProfileItem(self._state, self, items.profile_modifier) if items.profile_modifier else None,
        )

    async def profile_customisation_info(self, *, language: Language | None = None) -> ProfileCustomisation:
        """Fetch a user's profile customisation information.

        Parameters
        ----------
        language
            The language to fetch the profile items in. If ``None`` the current language is used
        """
        info = await self._state.fetch_user_profile_customisation(self.id64, language)
        return ProfileCustomisation(self._state, self, info)

    async def profile(self, *, language: Language | None = None) -> Profile:
        """Fetch a user's entire profile information.

        Parameters
        ----------
        language
            The language to fetch the profile items in. If ``None`` the current language is used

        Note
        ----
        This calls all the profile related functions to return a Profile object which has all the info set.
        """

        return Profile(
            *await asyncio.gather(
                self.equipped_profile_items(language=language),
                self.profile_customisation_info(language=language),
            )
        )

    def is_commentable(self) -> bool:
        """Specifies if the user's account can be commented on."""
        if hasattr(self, "is_friend"):
            return self.comment_permissions in (
                CommentPrivacyState.Public,
                CommentPrivacyState.FriendsOnly if self.is_friend() else CommentPrivacyState.Public,
            )
        return True  # our account

    def is_private(self) -> bool:
        """Specifies if the user has a private profile."""
        return self.privacy_state == CommunityVisibilityState.Private

    async def is_banned(self) -> bool:
        """Specifies if the user is banned from any part of Steam.

        Shorthand for:

        .. code-block:: python3

            bans = await user.bans()
            bans.is_banned()
        """
        bans = await self.bans()
        return bans.is_banned()

    def reviews(
        self,
        *,
        limit: int | None = None,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> UserReviewsIterator:
        """An :class:`~steam.iterators.AsyncIterator` for accessing a user's :class:`~steam.Review`\\s.

        Examples
        --------
        Usage:

        .. code-block:: python3

            async for review in user.reviews(limit=10):
                print(f"Author: {review.author} {'recommended' if review.recommend 'doesn\\'t recommend'} {review.game}")

        Flattening into a list:

        .. code-block:: python3

            reviews = await user.reviews(limit=50).flatten()
            # reviews is now a list of Review

        All parameters are optional.

        Parameters
        ----------
        limit
            The maximum number of reviews to search through. Setting this to ``None`` will fetch all the
            user's reviews.
        before
            A time to search for reviews before.
        after
            A time to search for reviews after.

        Yields
        ------
        :class:`~steam.Review`
        """
        return UserReviewsIterator(self._state, self, limit, before, after)

    async def fetch_review(self, game: Game) -> Review:
        """Fetch this user's review for a game.

        Parameters
        ----------
        game
            The games to fetch the reviews for.
        """
        (review,) = await self.fetch_reviews(game)
        return review

    async def fetch_reviews(self, *games: Game) -> list[Review]:
        """Fetch this user's review for games.

        Parameters
        ----------
        games
            The games to fetch the reviews for.
        """
        from .review import Review

        reviews = await self._state.fetch_user_review(self.id64, (game.id for game in games))
        return [Review._from_proto(self._state, review, self) for review in reviews]

    def published_files(
        self,
        *,
        game: Game | None = None,
        revision: PublishedFileRevision = PublishedFileRevision.Default,
        type: PublishedFileType = PublishedFileType.Community,
        language: Language | None = None,
        limit: int | None = None,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> UserPublishedFilesIterator:
        """An :class:`~steam.iterators.AsyncIterator` for accessing a user's :class:`~steam.PublishedFile`\\s.

        Examples
        --------

        Usage:

        .. code-block:: python3

            async for file in user.published_files(limit=10):
                print("Author:", file.author, "Published:", file.name)

        Flattening into a list:

        .. code-block:: python3

            files = await user.published_files(limit=50).flatten()
            # files is now a list of PublishedFile

        All parameters are optional.

        Parameters
        ----------
        game
            The game to fetch published files in.
        type
            The type of published file to fetch.
        revision
            The desired revision of the published file to fetch.
        language
            The language to fetch the published file in. If ``None``, the current language is used.
        limit
            The maximum number of published files to search through. Setting this to ``None`` will fetch all of the
            user's published files.
        before
            A time to search for published files before.
        after
            A time to search for published files after.

        Yields
        ------
        :class:`~steam.PublishedFile`
        """
        return UserPublishedFilesIterator(self._state, self, game, type, revision, language, limit, before, after)

    @classmethod
    def _patch_without_api(cls) -> None:
        import functools

        def __init__(self, state: ConnectionState, data: dict[str, Any]) -> None:
            super().__init__(data["steamid"])
            self._state = state
            self.name = data["persona_name"]
            self.avatar_url = data.get("avatar_url") or self.avatar_url

            self.real_name = NotImplemented
            self.trade_url = URL.COMMUNITY / f"tradeoffer/new?partner={self.id}"
            self.primary_clan = NotImplemented
            self.country = NotImplemented
            self.created_at = NotImplemented
            self.last_logoff = NotImplemented
            self.last_logon = NotImplemented
            self.last_seen_online = NotImplemented
            self.game = NotImplemented
            self.state = NotImplemented
            self.flags = NotImplemented
            self.privacy_state = NotImplemented
            self._commentable = NotImplemented
            self._setup_profile = NotImplemented
            self._level = data["level"]

        def __repr__(self) -> str:
            attrs = ("name", "id", "type", "universe", "instance")
            resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
            return f"<{self.__class__.__name__} {' '.join(resolved)}>"

        setattr(cls, "__init__", __init__)
        setattr(cls, "__repr__", __repr__)

        def not_implemented(function: str) -> None:
            @functools.wraps(getattr(cls, function))
            async def wrapped(*_, **__) -> None:
                raise NotImplementedError(
                    f"Accounts without an API key cannot use User.{function}, this is a Steam limitation not a library "
                    f"limitation, sorry."
                )

            setattr(cls, function, wrapped)

        not_implemented("friends")
        not_implemented("badges")
        not_implemented("is_commentable")
        not_implemented("is_private")
        not_implemented("is_banned")

    async def fetch_post(self, id: int) -> Post:
        ...

    def posts(self) -> AsyncIterator[Post]:
        ...


@runtime_checkable
class Messageable(Protocol[M_co]):
    """An ABC that details the common operations on a Steam message.
    The following classes implement this ABC:

        - :class:`~steam.User`
        - :class:`~steam.ClanChannel`
        - :class:`~steam.GroupChannel`
        - :class:`~steam.DMChannel`
    """

    __slots__ = ()

    @abc.abstractmethod
    def _message_func(self, content: str) -> Coroutine[Any, Any, M_co]:
        raise NotImplementedError

    @abc.abstractmethod
    def _image_func(self, image: Image) -> Coroutine[Any, Any, None]:
        raise NotImplementedError

    async def send(self, content: Any = None, image: Image | None = None) -> M_co | None:
        """Send a message to a certain destination.

        Parameters
        ----------
        content
            The content of the message to send.
        image
            The image to send to the user.

        Note
        ----
        Anything as passed to ``content`` is implicitly cast to a :class:`str`.

        Raises
        ------
        :exc:`~steam.HTTPException`
            Sending the message failed.
        :exc:`~steam.Forbidden`
            You do not have permission to send the message.

        Returns
        -------
        The sent message, only applicable if ``content`` is passed.
        """
        message = None if content is None else await self._message_func(str(content))
        if image is not None:
            await self._image_func(image)

        return message


@attr.dataclass(slots=True)
class Channel(Messageable[M_co]):
    _state: ConnectionState
    clan: Clan | None = None
    group: Group | None = None

    @abc.abstractmethod
    def history(
        self,
        *,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> AsyncIterator[M_co]:
        """An :class:`~steam.iterators.AsyncIterator` for accessing a channel's :class:`steam.Message`\\s.

        Examples
        --------

        Usage:

        .. code-block:: python3

            async for message in channel.history(limit=10):
                print("Author:", message.author, "Said:", message.content)

        Flattening into a list:

        .. code-block:: python3

            messages = await channel.history(limit=50).flatten()
            # messages is now a list of Message

        All parameters are optional.

        Parameters
        ----------
        limit
            The maximum number of messages to search through. Setting this to ``None`` will fetch all of the channel's
            messages, but this will be a very slow operation.
        before
            A time to search for messages before.
        after
            A time to search for messages after.

        Yields
        ------
        :class:`~steam.Message`
        """
        raise NotImplementedError


def _clean_up_content(content: str) -> str:  # steam does weird stuff with content
    return content.replace(r"\[", "[").replace("\\\\", "\\")


STEAM_EPOCH = datetime(2005, 1, 1, tzinfo=timezone.utc)


class Message(metaclass=abc.ABCMeta):
    """Represents a message from a :class:`~steam.User`. This is a base class from which all messages inherit.

    The following classes implement this:

        - :class:`~steam.UserMessage`
        - :class:`~steam.GroupMessage`
        - :class:`~steam.ClanMessage`

    .. container:: operations

        .. describe:: x == y

            Checks if two messages are equal

        .. describe:: hash(x)

            Returns the hash of a message.

    """

    __slots__ = (
        "author",
        "content",
        "channel",
        "clean_content",
        "created_at",
        "ordinal",
        "group",
        "clan",
        "mentions",
        "reactions",
        "partial_reactions",
        "_id_cs",
        "_state",
    )

    author: Authors
    """The message's author."""
    channel: Channel[Self]
    """The channel the message was sent in."""
    content: str
    """The message's content.

    Note
    ----
    This is **not** what you will see in the steam client see :attr:`clean_content` for that.
    """
    clean_content: str
    """The message's clean content without BBCode."""
    created_at: datetime
    """The time this message was sent at."""
    clan: Clan | None
    """The clan the message was sent in. Will be ``None`` if the message wasn't sent in a :class:`~steam.Clan`."""
    group: Group | None
    """The group the message was sent in. Will be ``None`` if the message wasn't sent in a :class:`~steam.Group`."""
    reactions: list[MessageReaction]
    """The message's reactions."""
    ordinal: int
    """A per-channel incremented integer up to ``1000`` for every message sent in a second window."""
    mentions: Mentions | None
    """An object representing mentions in this message."""
    reactions: list[MessageReaction]
    """The reactions this message has received."""
    partial_reactions: list[PartialMessageReaction]

    def __init__(self, channel: Channel[Self], proto: Any):
        self._state: ConnectionState = channel._state
        self.channel = channel
        self.group = channel.group
        self.clan = channel.clan
        self.content = _clean_up_content(proto.message)
        self.ordinal = proto.ordinal
        self.clean_content = getattr(proto, "message_no_bbcode", "") or self.content
        self.mentions = getattr(proto, "mentions", None)
        self.partial_reactions = []
        self.reactions = []

    def __repr__(self) -> str:
        attrs = ("author", "id", "channel")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    def __eq__(self, other: object) -> bool:
        return (
            self.channel == other.channel and self.id == other.id
            if isinstance(other, self.__class__)
            else NotImplemented
        )

    def __hash__(self) -> int:
        return hash((self.channel, self.id))

    @cached_slot_property
    def id(self) -> int:
        """A unique identifier for every message sent in a channel.

        Note
        ----
        This is **not** something Steam provides, this is meant to be a simple way to compare messages.
        """
        # a u64 "snowflake-esk" id measuring of the number of seconds passed since Steam's EPOCH and then the
        # "sequence"/ordinal of the message.
        return int(
            f"{int((self.created_at - STEAM_EPOCH).total_seconds()):032b}{self.ordinal:032b}",
            base=2,
        )

    # @abc.abstractmethod
    # async def delete(self) -> None:
    #     raise NotImplementedError()

    @abc.abstractmethod
    async def add_emoticon(self, emoticon: Emoticon) -> None:
        """Adds an emoticon to this message.

        Parameters
        ----------
        emoticon
            The emoticon to add to this message.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def remove_emoticon(self, emoticon: Emoticon) -> None:
        """Removes an emoticon from this message.

        Parameters
        ----------
        emoticon
            The emoticon to remove from this message.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def add_sticker(self, sticker: Sticker) -> None:
        """Adds a sticker to this message.

        Parameters
        ----------
        sticker
            The sticker to add to this message.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def remove_sticker(self, sticker: Sticker) -> None:
        """Adds a sticker to this message.

        Parameters
        ----------
        sticker
            The sticker to remove from this message.
        """
        raise NotImplementedError()



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import asyncio
import errno
import logging
import lzma
import struct
import sys
from base64 import b64decode
from collections.abc import AsyncGenerator, Generator
from contextlib import asynccontextmanager
from dataclasses import dataclass
from datetime import datetime
from io import BytesIO
from operator import attrgetter, methodcaller
from typing import TYPE_CHECKING, Any, Sequence, TypeVar, overload
from zipfile import BadZipFile, ZipFile
from zlib import crc32

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from multidict import MultiDict
from typing_extensions import Final, Never, Self
from yarl import URL

from . import utils
from ._const import VDF_LOADS, VDFDict
from .abc import SteamID
from .enums import AppFlag, BillingType, DepotFileFlag, Language, LicenseType, PackageStatus, ReviewType
from .game import StatefulGame
from .models import _IOMixin
from .package import StatefulPackage
from .protobufs import app_info
from .protobufs.content_manifest import Metadata, Payload, PayloadFileMapping, PayloadFileMappingChunkData, Signature
from .utils import DateTime

if sys.platform == "win32":
    from pathlib import PureWindowsPath as PurePathBase
else:
    from pathlib import PurePosixPath as PurePathBase


if TYPE_CHECKING:
    from .state import ConnectionState
    from .types import manifest
    from .types.vdf import VDFInt


__all__ = (
    "Manifest",
    "ManifestPath",
    "Branch",
    "ManifestInfo",
    "PrivateManifestInfo",
    "HeadlessDepot",
    "Depot",
    "GameInfo",
    "PackageInfo",
)

log = logging.getLogger(__name__)
NameT = TypeVar("NameT", bound="str | None", covariant=True)


def unzip(data: bytes) -> bytes:
    if data[:2] == b"VZ":
        if data[-2:] != b"zv":
            raise RuntimeError(f"VZ: Invalid footer: {data[-2:]!r}")
        if data[2:3] != b"a":
            raise RuntimeError(f"VZ: Invalid version: {data[2:3]!r}")

        filters = (lzma._decode_filter_properties(lzma.FILTER_LZMA1, data[7:12]),)  # type: ignore
        decompressor = lzma.LZMADecompressor(lzma.FORMAT_RAW, filters=filters)
        checksum, decompressed_size = struct.unpack("<II", data[-10:-2])
        data = decompressor.decompress(data[12:-9], max_length=decompressed_size)

        if crc32(data) != checksum:
            raise RuntimeError("VZ: CRC32 checksum doesn't match for decompressed data")

    else:
        try:
            with ZipFile(BytesIO(data)) as zf:
                data = zf.read(zf.filelist[0])
        except BadZipFile:
            pass

    return data


@dataclass
class ManifestPathParents(Sequence["ManifestPath"]):
    # this class is mostly a copy and paste from pathlib.py
    __slots__ = ("path",)
    path: ManifestPath  # we don't care about reference cycles

    def __len__(self) -> int:
        if self.path.drive or self.path.root:  # this should never be True but w/e
            return len(self.path.parts) - 1
        else:
            return len(self.path.parts)

    @overload
    def __getitem__(self, idx: int) -> ManifestPath:
        ...

    @overload
    def __getitem__(self, idx: slice) -> tuple[ManifestPath, ...]:
        ...

    def __getitem__(self, idx: int | slice) -> ManifestPath | tuple[ManifestPath, ...]:
        if isinstance(idx, slice):
            return tuple(self[i] for i in range(*idx.indices(len(self))))

        if idx >= len(self) or idx < -len(self):
            raise IndexError(idx)

        return self.path._from_parsed_parts(self.path.drive, self.path.root, self.path.parts[: -idx - 1])

    def __repr__(self) -> str:
        return f"<{self.path!r}.parents>"


class ManifestPath(PurePathBase, _IOMixin):
    """A :class:`pathlib.PurePath` subclass representing a binary file in a Manifest. This class is broadly compatible
    with :class:`pathlib.Path`.

    .. container:: operations

        .. describe:: x == y

            Checks if two paths are equal.

        .. describe:: hash(x)

            Hashes this path.

        .. describe:: x < y

            Checks if one path is less than the other.
    """

    __slots__ = ("_manifest", "_mapping", "_flags_cs")

    _manifest: Manifest
    _mapping: PayloadFileMapping

    def __new__(cls, manifest: Manifest, mapping: PayloadFileMapping) -> Self:
        # super().__new__ breaks
        self: Self = super()._from_parts(mapping.filename.rstrip("\x00 \n\t").split("\\"))  # type: ignore
        self._mapping = mapping
        self._manifest = manifest
        return self

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {str(self)!r}>"

    if not TYPE_CHECKING:

        def __getattr__(self, name: str) -> Never:
            if name in self.__annotations__:  # give a more helpful error
                raise AttributeError("Attempting operations on a non-existent file")
            raise AttributeError(f"{self.__class__.__name__!r} object has no attribute {name!r}")

    def _select_from_manifest(self, new_self: Self) -> Self:
        try:
            # try and return the actual path if exists
            return self._manifest._paths[new_self.parts]
        except KeyError:
            # else attach the manifest and return, this will not support most operations
            new_self._manifest = self._manifest
            return new_self

    def _from_parts(self, args: tuple[str, ...]) -> Self:
        new_self = super()._from_parts(args)  # type: ignore
        return self._select_from_manifest(new_self)

    def _from_parsed_parts(self, drv: str, root: str, parts: tuple[str, ...]) -> Self:
        new_self = super()._from_parsed_parts(drv, root, parts)  # type: ignore
        return self._select_from_manifest(new_self)

    @property
    def parents(self) -> ManifestPathParents:
        # cannot use the default implementation as it calls _from_parsed_parts as a classmethod
        return ManifestPathParents(self)

    @property
    def size(self) -> int:
        """This path's size in bytes."""
        return self._mapping.size

    @property
    def chunks(self) -> Sequence[PayloadFileMappingChunkData]:
        """A read-only version of this path's chunks instances."""
        return self._mapping.chunks

    @utils.cached_slot_property
    def flags(self) -> DepotFileFlag:
        """This path's flags."""
        return DepotFileFlag.try_value(self._mapping.flags)

    def is_dir(self) -> bool:
        """Whether the path is a directory."""
        return self.flags & DepotFileFlag.Directory > 0

    def is_symlink(self) -> bool:
        """Whether this path is a symlink."""
        return self.flags & DepotFileFlag.Symlink > 0

    def is_file(self) -> bool:
        """Whether this path is a file or symlink."""
        return not self.is_dir()

    def is_executable(self) -> bool:
        """Whether this file is executable."""
        return self.flags & DepotFileFlag.Executable > 0

    def is_hidden(self) -> bool:
        """Whether this file is hidden."""
        return self.flags & DepotFileFlag.Hidden > 0

    def readlink(self) -> Self:
        """If this path is a symlink where it points to. Similar to :meth:`pathlib.Path.readlink`

        Raises
        ------
        OSError
            this path isn't a symlink.
        """
        if not self.is_symlink():
            raise OSError(errno.EINVAL, f"Invalid argument: {str(self)!r}")

        link_parts = tuple(self._mapping.linktarget.rstrip("\x00 \n\t").split("\\"))
        return self._manifest._paths[link_parts]

    def iterdir(self) -> Generator[Self, None, None]:
        """Iterate over this path. Similar to :meth:`pathlib.Path.iterdir`."""
        for path in self._manifest._paths.values():
            if path.parent == self:
                yield path

    def glob(self, pattern: str) -> Generator[Self, None, None]:
        """Perform a glob operation on this path. Similar to :meth:`pathlib.Path.glob`."""
        if not pattern:
            raise ValueError(f"Unacceptable pattern: {pattern!r}")

        # str.removeprefix
        yield from filter(
            methodcaller("match", f"{self.as_posix().rstrip('/')}/{pattern}"), self._manifest._paths.values()
        )

    def rglob(self, pattern: str) -> Generator[Self, None, None]:
        """Perform a recursive glob operation on this path. Similar to :meth:`pathlib.Path.rglob`."""
        yield from self.glob(f"**/{pattern}")

    @asynccontextmanager
    async def open(self) -> AsyncGenerator[BytesIO, None]:
        """Reads the entire contents of the file into a :class:`io.BytesIO` object.

        Raises
        ------
        IsADirectoryError
            This path is a directory and cannot be opened.
        RuntimeError
            The depot cannot be decrypted as no key for its manifest was found.
        """
        if self.is_dir():
            raise IsADirectoryError(errno.EISDIR, f"Is a directory: {str(self)!r}")

        key = self._manifest._key
        if key is None:
            raise RuntimeError("Cannot decrypt this depot as we have no key.")

        with BytesIO() as buffer:
            for resp in await asyncio.gather(
                *(
                    self._manifest.server.get(f"depot/{self._manifest.depot_id}/chunk/{chunk.sha.hex()}")
                    for chunk in self.chunks
                )
            ):
                data = utils.symmetric_decrypt(resp, key)
                buffer.write(unzip(data))

            buffer.seek(0)
            yield buffer

    read_bytes = _IOMixin.read
    """Read the contents of the file. Similar to :meth:`pathlib.Path.read_bytes`"""

    async def read(self) -> Never:
        raise NotImplementedError("use read_bytes() instead of read()")

    async def read_text(self, encoding: str | None = None) -> str:
        """Read the contents of the file. Similar to :meth:`pathlib.Path.read_text`"""
        contents = await self.read_bytes()
        return contents.decode() if encoding is None else contents.decode(encoding)

    async def read_vdf(self, encoding: str | None = None) -> VDFDict:
        """Read the contents of the file into a VDFDict."""
        return VDF_LOADS(await self.read_text(encoding))


PAYLOAD_MAGIC: Final = 0x71F617D0
METADATA_MAGIC: Final = 0x1F4812BE
SIGNATURE_MAGIC: Final = 0x1B81B817
END_OF_MANIFEST_MAGIC: Final = 0x32C415AB


class Manifest:
    """Represents a manifest which is a collection of files included with a depot build on Steam's CDN.

    .. container:: operations

        .. describe:: x == y

            Checks if two manifests are equal.

        .. describe:: len(x)

            Returns the number of files this manifest holds.

    Attributes
    ----------
    name
        The name of the manifest.
    game
        The game that this manifest was fetched from.
    created_at
        The time at which the depot was created at.
    """

    __slots__ = (
        "name",
        "game",
        "server",
        "created_at",
        "_key",
        "_metadata",
        "_payload",
        "_signature",
        "_paths",
        "_state",
    )

    def __init__(self, state: ConnectionState, server: ContentServer, game_id: int, data: bytes):
        self._state = state
        self.name: str | None = None
        self.game = StatefulGame(state, id=game_id)
        self.server = server
        self._key: bytes | None = None
        self._paths: dict[tuple[str, ...], ManifestPath] = {}

        with utils.StructIO(unzip(data)) as io:
            if io.read_u32() != PAYLOAD_MAGIC:
                raise RuntimeError("Expecting protobuf payload")
            length = io.read_u32()
            self._payload = Payload().parse(io.read(length))

            if io.read_u32() != METADATA_MAGIC:
                raise RuntimeError("Expecting protobuf metadata")
            length = io.read_u32()
            self._metadata = Metadata().parse(io.read(length))

            if io.read_u32() != SIGNATURE_MAGIC:
                raise RuntimeError("Expecting protobuf signature")
            length = io.read_u32()
            self._signature = Signature().parse(io.read(length))

            if io.read_u32() != END_OF_MANIFEST_MAGIC:
                raise RuntimeError("Expecting end of manifest")

        self.created_at = DateTime.from_timestamp(self._metadata.creation_time)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r} id={self.id} depot_id={self.depot_id}>"

    def __eq__(self, other: object) -> bool:
        return self.id == other.id if isinstance(other, self.__class__) else NotImplemented

    def __len__(self) -> int:
        return len(self._payload.mappings)

    @property
    def paths(self) -> Sequence[ManifestPath]:
        """The depot's files."""
        if self._paths:
            return list(self._paths.values())

        paths = self._paths
        for mapping in self._payload.mappings:
            path = ManifestPath(self, mapping)
            self._paths[path.parts] = path
        return list(paths.values())

    @property
    def id(self) -> int:
        """The ID of this manifest. This is randomly generated by Steam."""
        return self._metadata.gid_manifest

    @property
    def depot_id(self) -> int:
        """The ID of this manifest's depot."""
        return self._metadata.depot_id

    @property
    def size_original(self) -> int:
        """The size of the original depot file."""
        return self._metadata.cb_disk_original

    @property
    def size_compressed(self) -> int:
        """The size of the compressed depot file."""
        return self._metadata.cb_disk_compressed

    @property
    def compressed(self) -> bool:
        """Whether the depot is compressed."""
        return self.size_original != self.size_compressed if self.size_compressed else False


@dataclass
class ContentServer(SteamID):  # is there any point having this inherit steamid?
    _state: ConnectionState
    url: URL
    weighted_load: float

    async def get(self, path: str) -> bytes:
        async with self._state.http._session.get(self.url / path) as r:
            return await r.read()

    async def fetch_manifest(
        self,
        game_id: int,
        id: int,
        depot_id: int,
        name: str | None = None,
        branch: str = "public",
        password_hash: str = "",
    ) -> Manifest:
        branch = branch if branch != "public" else ""
        code = await self._state.fetch_manifest_request_code(id, depot_id, game_id, branch, password_hash)
        data = await self.get(f"depot/{depot_id}/manifest/{id}/5{f'/{code}' if code else ''}")

        manifest = Manifest(self._state, self, game_id, data)
        encrypted = manifest._metadata.filenames_encrypted
        if encrypted:
            key = manifest._key = await self._state.fetch_depot_key(game_id, depot_id)
        for mapping in manifest._payload.mappings:
            if encrypted:
                mapping.filename = utils.symmetric_decrypt(b64decode(mapping.filename), key).decode()  # type: ignore # key is never unbound
                if mapping.linktarget:
                    mapping.linktarget = utils.symmetric_decrypt(b64decode(mapping.linktarget), key).decode()  # type: ignore

            mapping.chunks.sort(key=attrgetter("offset"), reverse=False)

        manifest.name = name
        return manifest


class Branch:
    """Represents a branch on for a Steam app. Branches are specific builds of an application that have made available
    publicly or privately through Steam.

    Read more on `steamworks <https://partner.steamgames.com/doc/store/application/branches>`_.

    Attributes
    ----------
    name
        The name of the branch.
    build_id
        The branch's build ID. This is a globally incrementing number. Build IDs are updated when a new build of an
        application is pushed.
    password_required
        Whether a password is required to access this branch.
    updated_at
        The time this branch was last updated. This can be ``None`` if the branch is ancient and hasn't been updated
        in ages.
    description
        This branch's description.
    depots
        This branch's depots.
    password
        This branch's password.
    """

    __slots__ = (
        "name",
        "build_id",
        "password_required",
        "updated_at",
        "description",
        "depots",
        "password",
    )

    def __init__(
        self,
        name: str,
        build_id: int,
        updated_at: datetime | None,
        password_required: bool,
        description: str | None,
    ):
        self.name = name
        self.build_id = build_id
        self.password_required = password_required
        self.updated_at = updated_at
        self.description = description
        self.depots: list[Depot] = []
        self.password: str | None = None

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r} build_id={self.build_id}>"

    @property
    def manifests(self) -> list[ManifestInfo]:
        """This branch's manifests."""
        return [depot.manifest for depot in self.depots]

    async def fetch_manifests(self) -> list[Manifest]:
        """Fetch this branch's manifests. Similar to :meth:`StatefulGame.manifests`."""
        return await asyncio.gather(*(manifest.fetch() for manifest in self.manifests))  # type: ignore  # typeshed lies


class ManifestInfo:
    """Represents information about a manifest.

    Attributes
    ----------
    id
        The manifest's ID.
    branch
        The branch this manifest is for.
    depot
        The depot this manifest is for.
    """

    __slots__ = ("_state", "id", "branch", "depot")
    depot: Depot

    def __init__(
        self,
        state: ConnectionState,
        id: int,
        branch: Branch,
    ):
        self._state = state
        self.id = id
        self.branch = branch

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r} id={self.id}>"

    @property
    def name(self) -> str | None:
        """The manifest's name."""
        return self.depot.name

    async def fetch(self) -> Manifest:
        """Resolves this manifest info into a full :class:`Manifest`."""
        return await self._state.fetch_manifest(self.depot.game.id, self.id, self.depot.id, self.name)


class PrivateManifestInfo(ManifestInfo):
    __slots__ = ("encrypted_id",)

    def __init__(self, state: ConnectionState, encrypted_id: str, branch: Branch):
        self._state = state
        self.encrypted_id = encrypted_id
        self.branch = branch

    @property
    def id(self) -> int:
        if self.branch.password is None:
            raise ValueError("Cannot access the id of this depot as the password is not set.")
        cipher = Cipher(algorithms.AES(self.branch.password.encode("UTF-8")), modes.ECB())
        decryptor = cipher.decryptor()
        to_unpack = utils.unpad(decryptor.update(bytes.fromhex(self.encrypted_id) + decryptor.finalize()))
        return struct.unpack("<Q", to_unpack)[0]

    @staticmethod
    def _get_id(depots: manifest.Depot, branch: Branch) -> VDFInt | None:
        try:
            return depots["encryptedmanifests"][branch.name]["encrypted_gid_2"]
        except KeyError:
            return None


@dataclass
class HeadlessDepot:
    """Represents a depot without a branch."""

    __slots__ = ("id", "name", "game", "max_size", "config", "shared_install", "system_defined")
    id: int
    """The depot's ID."""
    name: str
    """The depot's name."""
    game: GameInfo
    """The depot's game."""
    max_size: int
    """The depot's maximum size."""
    config: MultiDict[str]
    """The depot's configuration settings."""
    shared_install: bool
    """Whether this depot supports shared installs"""
    system_defined: bool
    """Whether this depot is system defined."""

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r} id={self.id} game={self.game!r}>"

    def __eq__(self, other: object) -> bool:
        return self.id == other.id if isinstance(other, self.__class__) else NotImplemented


@dataclass
class Depot(HeadlessDepot):
    """Represents a depot which is a grouping of files.

    Read more on `steamworks <https://partner.steamgames.com/doc/store/application/depots>`_.
    """

    __slots__ = ("manifest", "branch")
    manifest: ManifestInfo
    """The depot's associated manifest."""
    branch: Branch
    """The branch the depot is for."""


class ProductInfo:
    __slots__ = ()
    _state: ConnectionState
    sha: str
    size: int
    change_number: int

    def __init__(
        self,
        state: ConnectionState,
        proto: (
            app_info.CMsgClientPicsProductInfoResponseAppInfo | app_info.CMsgClientPicsProductInfoResponsePackageInfo
        ),
        **kwargs: Any,
    ):
        super().__init__(state, **kwargs)  # type: ignore  # this is a mixin
        self.sha = proto.sha.hex()
        self.size = proto.size
        self.change_number = proto.change_number

    # async def changes(self) -> ...:
    #     """A method to fetch the changes to this game since this change number"""
    #     changes = await self._state.fetch_changes_since(self.change_number, True, True)
    #     return changes.app_changes[0].change_number


class GameInfo(ProductInfo, StatefulGame):
    """Represents a collection of information on a game.

    Attributes
    ----------
    type
        The game's type.
    has_adult_content
        Whether this game has adult content according to Steam.
    has_adult_content_violence
        Whether this game has adult violence according to Steam.
    market_presence
        Whether this game has a market presence.
    workshop_visible
        Whether this game has a market presence.
    community_hub_visible
        Whether this game has a content hub visible.
    controller_support
        This game's level of controller support.
    publishers
        This game's publishers.
    developers
        This game's developers.
    supported_languages
        This game's supported languages.
    created_at
        The time this game was created.
    review_score
        This game's review score.
    review_percentage
        This game's review percentage.
    partial_dlc
        This game's downloadable content.
    icon_url
        This game's icon URL.
    logo_url
        This game's logo URL.
    website_url
        This game's URL.
    headless_depots
        The depots for this game without a branch.
    sha
        The game's SHA for this product info.
    size
        The product info's size.
    change_number
        The product info's change number.
    """

    __slots__ = (
        "sha",
        "size",
        "change_number",
        "_branches",
        "headless_depots",
        "type",
        "has_adult_content",
        "has_adult_content_violence",
        "market_presence",
        "workshop_visible",
        "community_hub_visible",
        "controller_support",
        "publishers",
        "developers",
        "supported_languages",
        "created_at",
        "review_score",
        "review_percentage",
        "partial_dlc",
        "icon_url",
        "logo_url",
        "website_url",
        "parent",
        "_stats_visible",
        "_free",
        "_on_windows",
        "_on_mac_os",
        "_on_linux",
    )
    name: str

    def __init__(
        self,
        state: ConnectionState,
        data: manifest.GameInfo,
        proto: app_info.CMsgClientPicsProductInfoResponseAppInfo,
    ):
        common = data["common"]
        extended = data.get("extended", {})
        super().__init__(state, proto, id=proto.appid, name=common["name"])
        self.type = AppFlag.from_str(common["type"])
        self.has_adult_content = common.get("has_adult_content", "0") == "1"
        self.has_adult_content_violence = common.get("has_adult_content_violence", "0") == "1"
        self.market_presence = common.get("market_presence", "0") == "1"
        self.workshop_visible = common.get("workshop_visible", "0") == "1"
        self.community_hub_visible = common.get("community_hub_visible", "0") == "1"
        self._stats_visible = common.get("community_visible_stats", "0") == "1"
        self._free = extended.get("isfreeapp", "0") == "1"

        self.controller_support = common.get("controller_support", "none")

        self.publishers: list[str] = [
            publisher["name"]
            for publisher in common.get("associations", {}).values()
            if publisher["type"] == "publisher"
        ]
        self.developers: list[str] = [
            developer["name"]
            for developer in common.get("associations", {}).values()
            if developer["type"] == "developer"
        ]
        self.supported_languages: list[Language] = [
            Language.from_str(language) for language, value in common.get("languages", {}).items() if value == "1"
        ]

        # TODO categories/genres
        self.created_at = (
            DateTime.from_timestamp(int(common["steam_release_date"])) if "steam_release_date" in common else None
        )
        self.review_score = ReviewType.try_value(int(common.get("review_score", 0)))
        self.review_percentage = int(common.get("review_percentage", 0))
        dlc = extended.get("listofdlc", "")
        self.partial_dlc = [StatefulGame(state, id=int(id)) for id in dlc.split(",")] if dlc else []

        os_list = common.get("oslist", "")
        self._on_windows = "windows" in os_list
        self._on_mac_os = "macos" in os_list
        self._on_linux = "linux" in os_list

        self.icon_url = (
            f"https://media.steampowered.com/steamcommunity/public/images/apps/{self.id}/{common['icon']}.jpg"
            if "icon" in common
            else None
        )
        self.logo_url = (
            f"https://media.steampowered.com/steamcommunity/public/images/apps/{self.id}/{common['logo']}.jpg"
            if "logo" in common
            else None
        )
        self.website_url = extended.get("homepage")
        self.parent = StatefulGame(state, id=int(common["parent"])) if "parent" in common else None

        depots: manifest.Depot = data.get("depots", {})  # type: ignore
        self._branches: dict[str, Branch] = {}

        for name, value in depots.get("branches", {}).items():
            try:
                build_id = int(value["buildid"])
            except KeyError:
                log.debug("Got a branch %s with no build id, discarding", name)
            else:
                self._branches[name] = Branch(
                    name=name,
                    build_id=build_id,
                    updated_at=DateTime.from_timestamp(int(value["timeupdated"])) if "timeupdated" in value else None,
                    password_required=bool(int(value.get("pwdrequired", False))),
                    description=value.get("description") or None,
                )
        self.headless_depots: Sequence[HeadlessDepot] = []

        for key, depot in depots.items():  # type: ignore
            try:
                id = int(key)
            except ValueError:
                continue
            else:  # only the int keys have VDFDicts
                depot: manifest.Depot
                kwargs = {
                    "id": id,
                    "name": depot.get("name"),
                    "config": depot.get("config", MultiDict()),
                    "max_size": int(depot["maxsize"]) if "maxsize" in depot else None,
                    "game": self,
                    "shared_install": bool(int(depot.get("sharedinstall", False))),
                    "system_defined": bool(int(depot.get("system_defined", False))),
                }
                try:
                    manifests = depot["manifests"]
                except KeyError:
                    self.headless_depots.append(HeadlessDepot(**kwargs))  # type: ignore
                else:
                    for branch_name, manifest_id in manifests.items():
                        branch = self._branches[branch_name]
                        if not branch.password_required:
                            manifest = ManifestInfo(state, int(manifest_id), branch=branch)
                        else:
                            encrypted_id = PrivateManifestInfo._get_id(depot, branch)
                            if encrypted_id is not None:
                                manifest = PrivateManifestInfo(state, encrypted_id, branch)
                            else:  # fall back to the public version
                                manifest = ManifestInfo(
                                    state,
                                    int(manifests["public"]),
                                    branch=self.public_branch,
                                )
                        depot_ = Depot(**kwargs, branch=branch, manifest=manifest)  # type: ignore
                        manifest.depot = depot_
                        branch.depots.append(depot_)

    def get_branch(self, name: str) -> Branch | None:
        """Get a branch by its name."""
        return self._branches.get(name)

    @property
    def branches(self) -> Sequence[Branch]:
        """The branches for this game."""
        return list(self._branches.values())

    @property
    def public_branch(self) -> Branch:
        """The public branch. Shorthand for:

        .. code-block:: python3

            game.get_branch("public")
        """
        return self._branches["public"]

    async def depots(self) -> Sequence[Depot | HeadlessDepot]:
        return [depot for branch in self._branches.values() for depot in branch.depots] + self.headless_depots  # type: ignore

    def is_on_windows(self) -> bool:
        """Whether the game is playable on Windows."""
        return self._on_windows

    def is_on_mac_os(self) -> bool:
        """Whether the game is playable on macOS."""
        return self._on_mac_os

    def is_on_linux(self) -> bool:
        """Whether the game is playable on Linux."""
        return self._on_linux

    def has_visible_stats(self) -> bool:
        """Whether the game has publicly visible stats."""
        return self._stats_visible

    def is_free(self) -> bool:
        """Whether the game is free to download."""
        return self._free

    def __repr__(self) -> str:
        attrs = ("name", "id", "type", "sha", "change_number")
        resolved = [f"{name}={getattr(self, name)!r}" for name in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"


class PackageInfo(ProductInfo, StatefulPackage):
    """Represents a collection of information on a package.

    Attributes
    ----------
    games
        The games included in the package.
    billing_type
        The billing type for the package.
    license_type
        The license type for the package.
    status
        The status for the package.
    depot_ids
        The depot IDs in of the depots in the package.
    app_items
        The app items in the package.
    """

    __slots__ = (
        "sha",
        "size",
        "change_number",
        "_games",
        "billing_type",
        "license_type",
        "status",
        "depot_ids",
        "app_items",
    )

    def __init__(
        self,
        state: ConnectionState,
        data: manifest.PackageInfo,
        proto: app_info.CMsgClientPicsProductInfoResponsePackageInfo,
    ):
        super().__init__(state, proto, id=proto.packageid)
        self._games = [StatefulGame(state, id=id) for id in data["appids"].values()]
        self.billing_type = BillingType.try_value(data["billingtype"])
        self.license_type = LicenseType.try_value(data["licensetype"])
        self.status = PackageStatus.try_value(data["status"])
        self.depot_ids = list(data["depotids"].values())
        self.app_items = list(data["appitems"].values())

    def __repr__(self) -> str:
        attrs = ("id", "sha", "change_number")
        resolved = [f"{name}={getattr(self, name)!r}" for name in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    async def games(self) -> list[StatefulGame]:
        return self._games



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import asyncio
import sys
from collections.abc import Coroutine
from datetime import timedelta
from operator import attrgetter
from typing import TYPE_CHECKING, Any

from . import utils
from ._const import URL
from .abc import BaseUser, Messageable
from .enums import CommunityVisibilityState, Language, PersonaState, PersonaStateFlag, Result, TradeOfferState, Type
from .errors import ClientException, ConfirmationError, HTTPException
from .game import Game, StatefulGame
from .profile import ClientUserProfile, OwnedProfileItems, ProfileInfo, ProfileItem
from .trade import Inventory, TradeOffer
from .utils import DateTime

if TYPE_CHECKING:
    from .friend import Friend
    from .image import Image
    from .message import UserMessage
    from .state import ConnectionState
    from .types import user
    from .types.id import ID64

__all__ = (
    "User",
    "ClientUser",
)


class _BaseUser(BaseUser):
    __slots__ = ()

    def __init__(self, state: ConnectionState, data: user.User):
        super().__init__(data["steamid"])
        self._state = state
        self.real_name = None
        self.community_url = None
        self.avatar_url = None
        self.primary_clan = None
        self.country = None
        self.created_at = None
        self.last_logoff = None
        self.last_logon = None
        self.last_seen_online = None
        self.game = None
        self.state = None
        self.flags = None
        self.privacy_state = None
        self.comment_permissions = None
        self.profile_state = None
        self.rich_presence = None
        self._update(data)

    def _update(self, data: user.User) -> None:
        self.name = data["personaname"]
        self.real_name = data.get("realname") or self.real_name
        self.community_url = data.get("profileurl") or super().community_url
        self.avatar_url = data.get("avatarfull") or self.avatar_url
        self.trade_url = URL.COMMUNITY / f"tradeoffer/new/?partner={self.id}"
        from .clan import Clan  # circular import

        self.primary_clan = (
            Clan(self._state, data["primaryclanid"]) if "primaryclanid" in data else self.primary_clan  # type: ignore
        )
        self.country = data.get("loccountrycode") or self.country
        self.created_at = DateTime.from_timestamp(data["timecreated"]) if "timecreated" in data else self.created_at
        self.last_logoff = DateTime.from_timestamp(data["lastlogoff"]) if "lastlogoff" in data else self.last_logoff
        self.last_logon = DateTime.from_timestamp(data["last_logon"]) if "last_logon" in data else self.last_logon
        self.last_seen_online = (
            DateTime.from_timestamp(data["last_seen_online"]) if "last_seen_online" in data else self.last_seen_online
        )
        self.rich_presence = data["rich_presence"] if "rich_presence" in data else self.rich_presence
        self.game = (
            StatefulGame(self._state, name=data["gameextrainfo"], id=data["gameid"]) if "gameid" in data else self.game
        )
        self.state = PersonaState.try_value(data.get("personastate", 0)) or self.state
        self.flags = PersonaStateFlag.try_value(data.get("personastateflags", 0)) or self.flags
        self.privacy_state = CommunityVisibilityState.try_value(data.get("communityvisibilitystate", 0))
        self.comment_permissions = CommunityVisibilityState.try_value(
            data.get("commentpermission", 0)
        )  # FIXME CommentPrivacyState
        self.profile_state = CommunityVisibilityState.try_value(data.get("profilestate", 0))


class User(_BaseUser, Messageable["UserMessage"]):
    """Represents a Steam user's account.

    .. container:: operations

        .. describe:: x == y

            Checks if two users are equal.

        .. describe:: str(x)

            Returns the user's name.

    Attributes
    ----------
    name
        The user's username.
    state
        The current persona state of the account (e.g. LookingToTrade).
    game
        The Game instance attached to the user. Is ``None`` if the user isn't in a game or one that is recognised by the
        api.
    avatar_url
        The avatar url of the user. Uses the large (184x184 px) image url.
    real_name
        The user's real name defined by them. Could be ``None``.
    primary_clan
        The user's primary clan.
    created_at
        The time at which the user's account was created. Could be ``None``.
    last_logon
        The last time the user logged into steam. This is only ``None`` if user hasn't been updated from the websocket.
    last_logoff
        The last time the user logged off from steam. Could be ``None`` (e.g. if they are currently online).
    last_seen_online
        The last time the user could be seen online. This is only ``None`` if user hasn't been updated from the
        websocket.
    country
        The country code of the account. Could be ``None``.
    flags
        The persona state flags of the account.
    """

    __slots__ = ()

    async def add(self) -> None:
        """Sends a friend invite to the user to your friends list."""
        await self._state.http.add_user(self.id64)

    async def remove(self) -> None:
        """Remove the user from your friends list."""
        await self._state.http.remove_user(self.id64)
        self._state.user._friends.pop(self.id64, None)

    async def cancel_invite(self) -> None:
        """Cancels an invitation sent to the user. This effectively does the same thing as :meth:`remove`."""
        await self._state.http.remove_user(self.id64)

    async def block(self) -> None:
        """Blocks the user."""
        await self._state.http.block_user(self.id64)

    async def unblock(self) -> None:
        """Unblocks the user."""
        await self._state.http.unblock_user(self.id64)

    async def escrow(self, token: str | None = None) -> timedelta | None:
        """Check how long any received items would take to arrive. ``None`` if the user has no escrow or has a
        private inventory.

        Parameters
        ----------
        token
            The user's trade offer token, not required if you are friends with the user.
        """
        resp = await self._state.http.get_user_escrow(self.id64, token)
        their_escrow = resp["response"].get("their_escrow")
        if their_escrow is None:  # private
            return None
        seconds = their_escrow["escrow_end_duration_seconds"]
        return timedelta(seconds=seconds) if seconds else None

    def _message_func(self, content: str) -> Coroutine[Any, Any, UserMessage]:
        return self._state.send_user_message(self.id64, content)

    def _image_func(self, image: Image) -> Coroutine[Any, Any, None]:
        return self._state.http.send_user_image(self.id64, image)

    async def send(
        self,
        content: Any = None,
        *,
        trade: TradeOffer | None = None,
        image: Image | None = None,
    ) -> UserMessage | None:
        """Send a message, trade or image to an :class:`User`.

        Parameters
        ----------
        content
            The message to send to the user.
        trade
            The trade offer to send to the user.

            Note
            ----
            This will have its :attr:`~steam.TradeOffer.id` attribute updated after being sent.

        image
            The image to send to the user.

        Raises
        ------
        :exc:`~steam.HTTPException`
            Sending the message failed.
        :exc:`~steam.Forbidden`
            You do not have permission to send the message.

        Returns
        -------
        The sent message only applicable if ``content`` is passed.
        """

        message = await super().send(content, image)
        if trade is not None:
            to_send = [item.to_dict() for item in trade.items_to_send]
            to_receive = [item.to_dict() for item in trade.items_to_receive]
            try:
                resp = await self._state.http.send_trade_offer(
                    self, to_send, to_receive, trade.token, trade.message or ""
                )
            except HTTPException as e:
                if e.code == Result.Revoked and (
                    any(item.owner != self for item in trade.items_to_receive)
                    or any(item.owner != self._state.user for item in trade.items_to_send)
                ):
                    if sys.version_info >= (3, 11):
                        e.add_note(
                            "You've probably sent an item isn't either in your inventory or the user's inventory"
                        )
                    else:
                        raise ValueError(
                            "You've probably sent an item isn't either in your inventory or the user's inventory"
                        ) from e
                raise e
            trade._has_been_sent = True
            needs_confirmation = resp.get("needs_mobile_confirmation", False)
            trade._update_from_send(self._state, resp, self, active=not needs_confirmation)
            if needs_confirmation:
                for tries in range(5):
                    try:
                        await trade.confirm()
                    except ConfirmationError:
                        break
                    except ClientException:
                        await asyncio.sleep(tries * 2)
                trade.state = TradeOfferState.Active

            # make sure the trade is updated before this function returns
            self._state._trades[trade.id] = trade
            self._state._trades_to_watch.add(trade.id)
            await self._state.wait_for_trade(trade.id)
            self._state.dispatch("trade_send", trade)

        return message

    def is_friend(self) -> bool:
        """Whether the user is in the :class:`ClientUser`'s friends."""
        return self.id64 in self._state.user._friends


class ClientUser(_BaseUser):
    """Represents your account.

    .. container:: operations

        .. describe:: x == y

            Checks if two users are equal.

        .. describe:: str(x)

            Returns the user's name.

    Attributes
    ----------
    name
        The user's username.
    friends
        A list of the :class:`ClientUser`'s friends.
    state
        The current persona state of the account (e.g. LookingToTrade).
    game
        The Game instance attached to the user. Is ``None`` if the user isn't in a game or one that is recognised by
        the api.
    avatar_url
        The avatar url of the user. Uses the large (184x184 px) image url.
    real_name
        The user's real name defined by them. Could be ``None``.
    primary_clan
        The user's primary clan. Could be ``None``
    created_at
        The time at which the user's account was created. Could be ``None``.
    last_logon
        The last time the user logged into steam. This is only ``None`` if user hasn't been updated from the websocket.
    last_logoff
        The last time the user logged off from steam. Could be ``None`` (e.g. if they are currently online).
    last_seen_online
        The last time the user could be seen online. This is only ``None`` if user hasn't been updated from the
        websocket.
    country
        The country code of the account. Could be ``None``.
    flags
        The persona state flags of the account.
    """

    # TODO more stuff to add https://github.com/DoctorMcKay/node-steamcommunity/blob/master/components/profile.js

    __slots__ = ("_friends",)

    def __init__(self, state: ConnectionState, data: user.User):
        super().__init__(state, data)
        self._friends: dict[ID64, Friend] = {}

    async def friends(self) -> list[Friend]:
        """Returns a list of the user's friends."""
        return list(self._friends.values())

    def get_friend(self, id: utils.Intable) -> Friend | None:
        """Get a friend from the client user's friends list."""
        id64 = utils.make_id64(id, type=Type.Individual)
        return self._friends.get(id64)

    async def inventory(self, game: Game, *, language: Language | None = None) -> Inventory:
        resp = await self._state.http.get_client_user_inventory(game.id, game.context_id, language)
        return Inventory(state=self._state, data=resp, owner=self, game=game, language=language)

    async def setup_profile(self) -> None:
        """Set up your profile if possible."""
        if self.has_setup_profile():
            return

        params = {"welcomed": 1}
        await self._state.http.get(URL.COMMUNITY / "my/edit", params=params)

    async def clear_nicks(self) -> None:
        """Clears the client user's nickname/alias history."""
        await self._state.http.clear_nickname_history()

    async def profile_items(self, *, language: Language | None = None) -> OwnedProfileItems:
        """Fetch all the client user's profile items.

        Parameters
        ----------
        language
            The language to fetch the profile items in. If ``None`` the current language is used
        """
        items = await self._state.fetch_profile_items(language)
        return OwnedProfileItems(
            backgrounds=[
                ProfileItem(self._state, self, background, um_name="ProfileBackground")
                for background in items.profile_backgrounds
            ],
            mini_profile_backgrounds=[
                ProfileItem(self._state, self, mini_profile_background, um_name="MiniProfileBackground")
                for mini_profile_background in items.mini_profile_backgrounds
            ],
            avatar_frames=[
                ProfileItem(self._state, self, avatar_frame, um_name="AvatarFrame")
                for avatar_frame in items.avatar_frames
            ],
            animated_avatars=[
                ProfileItem(self._state, self, animated_avatar, um_name="AnimatedAvatar")
                for animated_avatar in items.animated_avatars
            ],
            modifiers=[ProfileItem(self._state, self, modifier) for modifier in items.profile_modifiers],
        )

    async def profile(self, *, language: Language | None = None) -> ClientUserProfile:
        return ClientUserProfile(
            *await asyncio.gather(
                self.equipped_profile_items(language=language),
                self.profile_info(),
                self.profile_customisation_info(language=language),
                self.profile_items(language=language),
            )
        )

    async def profile_info(self) -> ProfileInfo:
        """The friend's profile info."""
        info = await self._state.fetch_friend_profile_info(self.id64)
        # why this is friend only I'm not really sure considering it's available through the API
        return ProfileInfo(
            created_at=DateTime.from_timestamp(info.time_created),
            real_name=info.real_name or None,
            city_name=info.city_name or None,
            state_name=info.state_name or None,
            country_name=info.country_name or None,
            headline=info.headline or None,
            summary=info.summary,
        )

    async def edit(
        self,
        *,
        name: str | None = None,
        real_name: str | None = None,
        url: str | None = None,
        summary: str | None = None,
        country: str | None = None,
        state: str | None = None,
        city: str | None = None,
        avatar: Image | None = None,
        # TODO privacy params, use ums
    ) -> None:
        """Edit the client user's profile. Any values that aren't set will use their defaults.

        Parameters
        ----------
        name
            The new name you wish to go by.
        real_name
            The real name you wish to go by.
        url
            The custom url ending/path you wish to use.
        summary
            The summary/description you wish to use.
        country
            The country you want to be from.
        state
            The state you want to be from.
        city
            The city you want to be from.
        avatar
            The avatar you wish to use.

            Note
            ----
            This needs to be at least 184px x 184px.

        Raises
        -------
        :exc:`~steam.HTTPException`
            Editing your profile failed.
        """
        if any((name, real_name, url, summary, country, state, city)):
            await self._state.http.edit_profile_info(name, real_name, url, summary, country, state, city)
        if avatar is not None:
            await self._state.http.update_avatar(avatar)
        # TODO privacy stuff


class WrapsUser(User if TYPE_CHECKING else BaseUser, Messageable["UserMessage"]):
    """Internal class used for creating a User subclass optimised for memory. Composes the original user and forwards
    all of its attributes.

    Similar concept to discord.py's Member except more generalised for the larger number situations Steam throws at us.
    Slightly different however in that isinstance(SubclassOfWrapsUser(), User) should pass.

    Note
    ----
    This class does not forward ClientUsers attribute's so things like Clan().me.clear_nicks() will fail.
    """

    __slots__ = ("_user",)

    def __init__(self, state: ConnectionState, user: User):
        super().__init__(user.id64, type=Type.Individual)  # type: ignore
        self._user = user

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        for name, function in set(User.__dict__.items()) - set(object.__dict__.items()):
            setattr(cls, name, function)
        for name in (*User.__slots__, *BaseUser.__slots__):
            setattr(cls, name, property(attrgetter(f"_user.{name}")))  # TODO time this with a compiled property
            # probably wont be different than the above

        User.register(cls)
